
#name LOADER

#autotype

COMMA			= ","
LB				= "("
RB				= ")"

$word			= {\D^\y}{\S^\y}*
$literal		= "(\\{^\n}|{^\\\n\"})*"

cline			= //
cbegin			= /{*}
cend			= {*}/

skip			= \s

preset default {
    +all
}

preset eoln {
	skip		= {\s^\n}
	$eoln		= \n
    +all
}

preset preset {
	"-"
    +$word +$literal +"skip" +"{" +"}" +"->" +"+" +"="
}

preset regexp {
	$regexp		= \S+
    -comment
}

preset name {
	$name		= {\S^\y}+
}

"#name" | "#start" <<< PushPreset(PRESET_EOLN) >>>
"preset"		<<< PreparePreset >>>
$regexp			|
$name	 		<<< PopPresetString >>>
$eoln  			<<< PopPreset >>>

$word, $literal, $name, $regexp : string

$literal		<<< ProcessLiteral >>>

program			: void
program			-> "#name" worl $eoln (declaration *);				<<< CommitGrammar >>>

declaration		-> "#start" worl $eoln;								<<< StartNonterminal >>>

worl			-> $word | $literal;
worl			-> "L" $literal;									<<< LiteralToString >>>

declaration		-> (param % "|") code_block;						<<< SetItemsCodeBlock >>>

code_block		: [CODE_BLOCK]
code_block	 	-> "<<<" code_name ">>>";							<<< SimpleCodeBlock >>>
code_block	 	-> "{"   code_name "}";								<<< SimpleCodeBlock >>>
code_block		-> "<<<" code_name params_ne ">>>"; 				<<< ParametrizedCodeBlock >>>
code_block		-> "{"   code_name params_ne "}"; 					<<< ParametrizedCodeBlock >>>

code_name		-> worl | "construct";

params_ne		-> LB (param % COMMA) RB;

declaration		-> rule ";" attrib_list;							<<< CommitRuleAndItsCodeBlock >>>

attrib_list		: [RULE_ATTRIB]
attrib_list		-> ;
attrib_list		-> attrib_list "shift" shift_params;				<<< ShiftAttrib >>>
attrib_list		-> attrib_list "reduce" shift_params;				<<< ReduceAttrib >>>
attrib_list		-> attrib_list "prior" "(" $word ")";				<<< PriorAttrib >>>
attrib_list		-> attrib_list code_block;							<<< CodeAttrib >>>

shift_params	-> [LB (shift_term +) RB];

shift_term		: [ITEM]
shift_term		-> $literal;										<<< CreateTermItem >>>
shift_term		-> $word;											<<< CreateWordItem(true) >>>
shift_term		-> "$eof";											<<< construct(TERMINAL, "TOKEN_EOF") >>>

rule			: [RULE]
rule			-> $word {FNT} "->" rule_list (nt_type -> [":" type]); <<< CreateRule >>>

rule_list		-> rule_item % "|";				  

rule_item		: [RULE_ITEM]
rule_item		-> ;												<<< construct("NULL") >>>
rule_item		-> item_list; shift("]")
rule_item		-> item_list "*";									<<< construct(PERIODIC) >>>
rule_item		-> item_list "+";									<<< construct(NEM_PERIODIC) >>>
rule_item		-> "[" item_list "]"; 								<<< construct(YESNOT) >>>
rule_item		-> item_list "%" item_list;

item_list		-> item +;

item			: [ITEM]
item			-> $literal;										<<< CreateTermItem >>>
item			-> $word;											<<< CreateWordItem(false) >>>
item			-> "$error";										<<< construct(TERMINAL, "TOKEN_ERROR") >>>

item			-> "(" rule ")";									<<< CreateBracketsItem >>>
item			-> "(" rule_list nt_type ")";						<<< CreateBracketsOrItem >>>

item			-> code_block;										<<< CreateCodeBlockItem >>>
item			-> "[" rule_item "]"; 								<<< CreateYesNotItem >>>

declaration		-> (item2 % ",") ":" type;							<<< RegisterResultType >>>

item2			: [ITEM]
item2			-> $word;											<<< CreateWordItem(false) >>>
item2			-> $literal;										<<< CreateTermItem >>>

type			: [TYPE]
type			-> "[" $word "]";									<<< Type(PTR, WORD)    >>>
type			-> "[" $literal "]";								<<< Type(PTR, LITERAL) >>>
type			-> $word;											<<< Type(VAL, WORD)    >>>
type			-> $literal;										<<< Type(VAL, LITERAL) >>>
type			-> $word "*";										<<< TypeArray(VAL, WORD)    >>>
type			-> $literal "*";									<<< TypeArray(VAL, LITERAL) >>>
type			-> "[" $word "]" "*";								<<< TypeArray(PTR, WORD)    >>>
type			-> "[" $literal "]" "*";							<<< TypeArray(PTR, LITERAL) >>>

param			: [PARAM]
declaration		-> "left"   (param +) ";";							<<< RegisterPriority(true) >>>
declaration		-> "right"  (param +) ";";							<<< RegisterPriority(false) >>>

param			-> $word;											<<< construct(WORD) >>>
param			-> $literal;										<<< construct(LITERAL) >>>
param			-> "L" $literal;									<<< construct(STRING) >>>

declaration		-> alias_name 	  "=" param;						<<< SetAlias >>>
declaration		-> alias_name 	  "=" $regexp regexp_prior; 		<<< LexRegexp >>>
alias_name		-> param;											<<< CheckAliasName >>>

regexp_prior	-> [{SPP} ":" $name]; // low, normal, high

declaration		-> "skip"	{ESP} "=" $regexp;						<<< LexSkip >>>
declaration		-> "cline"	{ESP} "=" $regexp;						<<< LexCLine >>>
declaration		-> "cbegin" {ESP} "=" $regexp;						<<< LexCBegin >>>
declaration		-> "cend"	{ESP} "=" $regexp;						<<< LexCEnd >>>

declaration		-> "preset" $name {PTNM} "{" (pr_com *) {PTND} "}";

pr_com			-> "+" param;										<<< LexCommand(true) >>>
pr_com			-> "-" param;										<<< LexCommand(false) >>>
pr_com			-> "skip" 	{ESP} "=" $regexp;						<<< LexPresetSkip >>>
pr_com			-> param 	{CPT} "=" $regexp regexp_prior;			<<< LexPresetRegexp >>>
pr_com			-> param;											<<< LexPresetTerm >>>

declaration 	-> "#autotype";										<<< EnableAutotype >>>

declaration		-> "msg" "{" (message *) "}";

message			-> $word "=" $literal ["," $literal] ";";			<<< Message >>>

msg {
	LDR1001		= "Terminal '{name}' hasn't got regular expression";
	LDR1002		= "Failed to build lex table";
	LDR1003		= "Unexpected nonterminal '{name}' in the terminal alias block";
    LDR1004		= "Unexpected terminal '{name}' in the rule declaration";
    LDR1005		= "Unexpected nonterminal '{name}' in the terminal list";
    LDR1006		= "Code block cannot be empty because custom terminal result";
    LDR1007		= "Code block cannot be empty because nonempty argument list";
    LDR1008		= "Unknown priority '{name}'";
    LDR1009		= "Warning, replacing code block for nonterminal '{name}'";
    LDR1010		= "Overriding priority";
	LDR1011		= "Prefer shift redefinition";
	LDR1012		= "Prefer reduce redefinition";
	LDR1013		= "Result type redefinition for terminal '{name}'";
	LDR1014		= "Result type redefinition for nonterminal '{name}'";
	LDR1015		= "Unable to declare array for elements with type 'void'";
	LDR1016		= "Overriding shift attribute";
	LDR1017		= "Overriding reduce attribute";
	LDR1018		= "Overriding code attribute '{name}'";
	LDR1019		= "Overriding priority attribute '{name}'";
	LDR1020		= "Overriding terminal in priority section '{name}'";
	LDR1021		= "Overriding priority alias '{name}'";
	LDR1022		= "Bad priority value '{name}', expected 'low', 'normal' or 'high'";
	LDR1023		= "Failed to compile regular expression '{name}'";
	LDR1024		= "Redefinition of 'skip' in preset declaration";
	LDR1025		= "Failed to compile line comment regexp";
	LDR1026		= "Unexpected nonterminal '{name}' in preset declaration";
	LDR1027		= "Unexpected nonterminal in preset's local terminal declaration";
	LDR1028		= "Name '{name}' reserved and can't be used as a preset name";
	LDR1029		= "Preset '{name}' is already defined";
	LDR1030		= "Unexpected token '{name}'";
	LDR1031		= "Basic type '{name}' should be declared as a value type";
	LDR1032		= "Value types do not allow constructors";
	LDR1033		= "Message with id '{name}' is already defined";
}
