
/*************************************************************************/
/* ECG.H                                                      09.06.2007 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_H
#define ECG_H

#pragma once

#pragma warning(push)
#pragma warning(disable : 4996)

#include "ecg_comm.h"
#include "ecg_ptr.h"
#include "ecg_stack.h"
#include "ecg_auto.h"
#include "ecg_message.h"

namespace ecg {

template <class LEX>
struct IOSTREAM_IO_POLICY {
	IOSTREAM_IO_POLICY(void)
		: input(), error(), buf(), bufCur(), remains()
	{}

	enum { BUFFER_SIZE = 4096 };

	bool Init(std::istream & input, std::ostream & error)
	{
		this->input		= &input;
		this->error		= &error;
		
		remains			= 0;
		if (buf == NULL) {
			buf = new char[BUFFER_SIZE];
		}
		return true;
	}

	inline bool GetChar(typename LEX::CHAR_T & ch)
	{
		if (remains == 0) {
			input->read(buf, BUFFER_SIZE);
			remains = input->gcount();
			if (remains == 0) {
				return false;
			}
			bufCur = buf;
		}

		ch = (typename LEX::CHAR_T)*bufCur++;
		--remains;
		return true;
	}

public:
	std::ostream &			GetError			(void) const { return *error; }

private:
	std::istream *			input;
	std::ostream *			error;
	char *					buf;
	char *					bufCur;
	int						remains;
};

template <class CHILD>
struct SIMPLE_ERROR_POLICY {
	void					Error				(void)
	{
		static_cast<CHILD *> (this)->GetError()
			<< "Line " << static_cast<CHILD *> (this)->Line() << ": Unexpected token \"" << static_cast<CHILD *> (this)->TokenValue() << "\"" << endl;
	}
};

template <class CHILD>
struct MESSAGE_ERROR_POLICY {
	void					Error				(void)
	{
		string token = "token \"" + TokenValue() + "\"";
		if (static_cast<CHILD *> (this)->Token() == TOKEN_EOF) {
			token = "end of file";
		}
		messageTable["name"] = TokenValue();
		PrintMessage(CHILD::MESSAGE_DEFAULT_ERROR);
	}

protected:
	void					PrintMessage		(int msgId, int line /* = -1 */)
	{
		std::string msg = GetMessageByIndex(msgId)->format;
		msg = "{info}({line}): {err}: " + msg;
		message["err"]  = GetMessageByIndex(msgId)->id;
		message["line"] = ToString((line == -1) ? Line() : line);
		static_cast<CHILD *> (this)->GetError()
			<< ecg::Compose(msg, message) << endl;
	}

protected:
	MESSAGE::STRING_MAP		message;
};

template <class CHILD, class IMPL, class LEX>
struct HELPER_CFG_DEFAULT {
	typedef IOSTREAM_IO_POLICY<LEX>		IO_POLICY;
	typedef SIMPLE_ERROR_POLICY<CHILD>	ERROR_POLICY;
};

struct EMPTY_ERROR_POLICY {};
struct EMPTY_IO_POLICY {};

template <class CHILD, class IMPL, class LEX>
struct HELPER_CFG_EMPTY_ERROR {
	typedef IOSTREAM_IO_POLICY<LEX>		IO_POLICY;
	typedef EMPTY_ERROR_POLICY			ERROR_POLICY;
};

template <class CHILD, class IMPL, class LEX>
struct HELPER_CFG_EMPTY {
	typedef EMPTY_IO_POLICY				IO_POLICY;
	typedef EMPTY_ERROR_POLICY			ERROR_POLICY;
};

template <class CHILD, class IMPL, class LEX, template <class, class, class> class CONFIG = HELPER_CFG_DEFAULT>
class HELPER : public IMPL, public LEX, public CONFIG<CHILD, IMPL, LEX>::IO_POLICY, public CONFIG<CHILD, IMPL, LEX>::ERROR_POLICY {
public:
	bool					HasErrors			(void) const { return compilerState.is(CST_ERROR_MASK); }

	void					StateOK				(void) { compilerState.clear(CST_CLEANUP | CST_FATAL_ERROR); }
	void					StateCleanUp		(void) { compilerState.set(CST_CLEANUP); }
	void					StateFatal			(void) { compilerState.set(CST_FATAL_ERROR); }
	bool					IsCleanUp			(void) const { return compilerState.is(CST_CLEANUP); }
};

}

#pragma warning(pop)

#endif // ECG_H
