
/*************************************************************************/
/* ECG_AUTO.H                                                 20.03.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_AUTO_H
#define ECG_AUTO_H

#ifndef TOKEN_EOF
#define TOKEN_EOF			-1
#endif

#ifndef TOKEN_ERROR
#define TOKEN_ERROR			-2
#endif

namespace ecg {

enum COMPILER_STATE {
	// by default all ok
	CST_DEFAULT				= 0x0000,

	// error states
	CST_ERROR_OCCURED		= 0x0001, // this flag is set up when at least one error has occurred
	CST_CLEANUP				= 0x0002, // set this to activate clean up
	CST_FATAL_ERROR			= 0x0004, // fatal error, no way to recover (stops compiling)
	
	CST_ERROR_MASK			= CST_ERROR_OCCURED | CST_CLEANUP | CST_FATAL_ERROR,

	// state for user to derive this flags
	CST_USER				= 0x0010,
};
DECL_FLAGS(COMPILER_STATE, CST_USER);

template <
	int STATE_COUNT_,
	int TERM_COUNT_,
	int NONTERM_COUNT_,
	int RULE_COUNT_
>
struct AUTO_CONFIG {
	enum {
		STATE_COUNT			= STATE_COUNT_,
		TERM_COUNT			= TERM_COUNT_,
		NONTERM_COUNT		= NONTERM_COUNT_,
		RULE_COUNT			= RULE_COUNT_,
	};
};

template <class T, class CONFIG>
class AUTO {
public:
	enum {
		STATE_COUNT			= CONFIG::STATE_COUNT,
		TERM_COUNT			= CONFIG::TERM_COUNT,
		NONTERM_COUNT		= CONFIG::NONTERM_COUNT,
		RULE_COUNT			= CONFIG::RULE_COUNT,
	};
	
	enum ACTION_TYPE {
		ERROR = 0,
		SHIFT,
		REDUCE,
	};

	struct ACTION {
		ACTION_TYPE		type;
		int				data; // if SHIFT - next state, if REDUCE - production number
	};
	
	typedef	void (T::* ACTION_PROC) ();

	struct PRODUCTION {
		ACTION_PROC		proc;
		int				stateCount; // state amount to pop from stack on production
		int				nonterm;
	};

	struct STATE {
		int				state;
		int				sstackSize;
		int				vstackSize;
		int				pstackSize;
	};
	typedef std::stack<STATE> STATE_STACK;

public:
	void				Compile					(T & context);

public:
	const ACTION 		(*actionTable)[STATE_COUNT][TERM_COUNT];
	const ACTION *		actionTableEof;
	const ACTION *		actionTableError;
	const int 			(*productionTable)[STATE_COUNT][NONTERM_COUNT];
	const PRODUCTION *	procTableProduction;
	const ACTION_PROC *	procTableTerm;
	const ACTION_PROC * procTableNonterm;

	STATE_STACK			lalrStack;
};

} // namespace ecg

#include "ecg_auto.hpp"

#endif // ECG_AUTO_H