
/*************************************************************************/
/* ECG_COMM.H                                                 28.08.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_COMM_H
#define ECG_COMM_H

#define ECG_VERSION 0x00010006

namespace ecg {

/*************************************************************************/
/* STATIC_ASSERT                                                         */
/*                                                                       */
/*************************************************************************/
template <int num>
struct INT_2_TYPE;

#define ECG_STATIC_ASSERT(COND) \
	void DummyAssertFunc(ecg::INT_2_TYPE<__COUNTER__>, int a[1 - 2 * !(COND)]);

#define ECG_NUM_ELEMENTS(ARR) \
	(sizeof(ARR) / sizeof(*(ARR)))

/*************************************************************************/
/* FLAGS                                                                 */
/*                                                                       */
/*************************************************************************/
template <class T, class C = int>
class FLAGS {
	typedef T			enum_t;
	typedef C			cont_t;
	cont_t				val;

public:
	FLAGS	 			(void) : val() {}

	template <class T>
	FLAGS				(T a) : val(a) { (void)sizeof(DummyFlagsFunc(static_cast<enum_t> (a), a)); }

	template <class T>
	inline void			set			(T a) { val |= a;	(void)sizeof(DummyFlagsFunc(static_cast<enum_t> (a), a)); }

	template <class T>
	inline void			set			(T a, bool set) { if (set) val |= a; else val &= ~a; (void)sizeof(DummyFlagsFunc(static_cast<enum_t> (a), a)); }

	template <class T>
	inline void			clear		(T a) { val &= ~a;	(void)sizeof(DummyFlagsFunc(static_cast<enum_t> (a), a)); }

	template <class T>
	inline void			toggle		(T a) { val ^= a;	(void)sizeof(DummyFlagsFunc(static_cast<enum_t> (a), a)); }

	template <class T>
	inline void			fill		(T a) { val = static_cast<cont_t> (a); (void)sizeof(DummyFlagsFunc(static_cast<enum_t> (a), a)); }

	inline void			clear		(void) { val = cont_t(); }

	inline enum_t		value		(void) const { return static_cast<enum_t> (val); }

	template <class T>
	inline bool			is			(T a) const { return (val & a) != cont_t(); (void)sizeof(DummyFlagsFunc(static_cast<enum_t> (a), a)); }

	template <class T>
	inline bool			isEach		(T a) const { return (val & a) == a;	(void)sizeof(DummyFlagsFunc(static_cast<enum_t> (a), a)); }

	template <class T>
	inline bool			isToggled	(ecg::FLAGS<enum_t, cont_t> a, T b) const { return (val & b) != (a.val & b);	(void)sizeof(DummyFlagsFunc(static_cast<enum_t> (b), b)); }
};


#define DECL_FLAGS_COMMON_S(NAME, NAME_ROOT)	\
	bool DummyFlagsFunc(NAME_ROOT, NAME);		\
	template <class T>							\
	NAME operator | (NAME a, T b) { return (NAME)((int)a | (int)b); (void)sizeof(DummyFlagsFunc(static_cast<NAME_ROOT> (a), b)); } \
	template <class T>							\
	NAME operator & (NAME a, T b) { return (NAME)((int)a & (int)b); (void)sizeof(DummyFlagsFunc(static_cast<NAME_ROOT> (a), b)); }

#define DECL_FLAGS_COMMON(NAME, LAST)			\
	enum { NAME##_IMPL_LAST = LAST };			\
	DECL_FLAGS_COMMON_S(NAME, NAME##_IMPL_ROOT);

// declare derivable flags
#define DECL_FLAGS(NAME, LAST)					\
	typedef ecg::FLAGS<NAME> NAME##_F;				\
	typedef NAME NAME##_IMPL_ROOT;				\
	DECL_FLAGS_COMMON(NAME, LAST);

// declare flags (S meane 'sealed' can't be derived)
#define DECL_FLAGS_S(NAME)						\
	typedef ecg::FLAGS<NAME> NAME##_F;				\
	DECL_FLAGS_COMMON_S(NAME, NAME);

// declare child flags (M means 'multiple' - allows to declare
// many flags with one root
#define DECL_CHILD_FLAGS_M(NAME, PARENT, LAST)	\
	typedef PARENT##_IMPL_ROOT NAME##_IMPL_ROOT; \
	STATIC_ASSERT(LAST >= PARENT##_IMPL_LAST);	\
	DECL_FLAGS_COMMON(NAME, LAST);

// declare unique child flags (can't be 
#define DECL_CHILD_FLAGS(NAME, PARENT, LAST)	\
	DECL_CHILD_FLAGS_M(NAME, PARENT, LAST);		\
	NAME DummyFlagsFunc(PARENT);

#define DECL_LOCAL_FLAGS(NAME, CLASS)			\
	typedef ecg::FLAGS<NAME> NAME##_F;				\
	friend NAME operator | (NAME a, NAME b) { return (CLASS::NAME)((int)a | (int)b); } \
	friend NAME DummyFlagsFunc(NAME a, NAME b);

} // namespace ecg

#endif // ECG_COMM_H
