
/*************************************************************************/
/* ECG_MESSAGE.H                                              22.10.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_MESSAGE_H
#define ECG_MESSAGE_H

namespace ecg {

/*************************************************************************/
/* struct MSG_HOLDER                                                     */
/*                                                                       */
/*************************************************************************/
struct MSG_HOLDER {
	const char *			id;
	const char *			format;
	const char *			info;

	bool operator < (const std::string & s) const { return id < s; }
	bool operator < (const MSG_HOLDER & p) const { return id < p.id; }
	friend bool operator < (const std::string & s, const MSG_HOLDER & p) { return s < p.id; }
};

/*************************************************************************/
/* class MESSAGE                                                         */
/*                                                                       */
/*************************************************************************/
class MESSAGE {
public:
	const std::string &		ID				(void) const { return id; }
	const std::string &		Format			(void) const { return format; }
	const std::string &		Info			(void) const { return info; }

	typedef std::map<std::string, std::string> STRING_MAP;

	std::string				Compose			(const STRING_MAP & string) const;

	MESSAGE					(const std::string & id, const std::string & format, const std::string & info)
		: id(id)
		, format(format)
		, info(info)
	{}

	MESSAGE					(const MSG_HOLDER & holder)
		: id		(holder.id		? holder.id		: "")
		, format	(holder.format	? holder.format	: "")
		, info		(holder.info	? holder.info	: "")
	{}

	MESSAGE					(const MESSAGE & msg)
		: id(msg.id)
		, format(msg.format)
		, info(msg.info)
	{}

private:
	std::string				id;				// type of message (user defined, like "W1" or "E" / "CE")
	std::string				format;			// format of message (like "Unexpected token '{token}'")
	std::string				info;			// user information for message (using {=my info})
};

int				FindMessage		(const std::string & id, const MSG_HOLDER * messageTable, int size);
std::string		Compose			(const std::string & format, const MESSAGE::STRING_MAP & table);
std::string		Compose			(const MSG_HOLDER * holder, const MESSAGE::STRING_MAP & table);

std::string		ToString		(int val);

#include "ecg_message.hpp"

} // namespace ecg

#endif // ECG_MESSAGE_H
