
/*************************************************************************/
/* ECG_MESSAGE.HPP                                            09.06.2007 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

/*************************************************************************/
/* MESSAGE::Compose()                                                    */
/*                                                                       */
/*************************************************************************/
inline std::string MESSAGE::Compose(const STRING_MAP & table) const
{
	return ecg::Compose(format, table);
}

/*************************************************************************/
/* Compose()                                                             */
/*                                                                       */
/*************************************************************************/
inline std::string Compose(const std::string & format, const MESSAGE::STRING_MAP & table)
{
	enum STATE {
		DEFAULT,
		DEFAULT_ESCAPE,
		TOKEN,
		TOKEN_ESCAPE,
	} state = DEFAULT;

	std::string res, name;
	res.reserve(format.size());
	int pos = 0;
	while (pos < (int)format.size()) {
		switch (state) {
		case DEFAULT:
			switch (format[pos]) {
				case '%':
					state = DEFAULT_ESCAPE;
					break;
				case '{':
					state = TOKEN;
					name.clear();
					break;
				default:
					res += format[pos];
					break;
			}
			break;

		case DEFAULT_ESCAPE:
			res += format[pos];
			state = DEFAULT;
			break;

		case TOKEN:
			switch (format[pos]) {
				case '%':
					state = TOKEN_ESCAPE;
					break;
				case '}':
					{
						MESSAGE::STRING_MAP::const_iterator iter = table.find(name);
						if (iter != table.end()) {
							res += iter->second;
						} else {
							res += "{" + name + "}";
						}
						state = DEFAULT;
					}
					break;
				default:
					name += format[pos];
					break;
			}
			break;

		case TOKEN_ESCAPE:
			name += format[pos];
			state = TOKEN;
			break;

		default:
			assert(false);
			break;
		}

		++pos;
	}

	return res;
}

/*************************************************************************/
/* ToString()                                                            */
/*                                                                       */
/*************************************************************************/
inline std::string ToString(int val)
{
	char buf[32];
	return itoa(val, buf, 10);
}

/*************************************************************************/
/* FindMessage()                                                         */
/*                                                                       */
/*************************************************************************/
inline int FindMessage(const std::string & id, const MSG_HOLDER * messageTable, int size)
{
	for (int i=0; i<size; ++i) {
		if (messageTable[i].id == id) {
			return i;
		}
	}

	return -1;
}

/*************************************************************************/
/* Compose()                                                             */
/*                                                                       */
/*************************************************************************/
inline std::string Compose(const MSG_HOLDER * holder, const MESSAGE::STRING_MAP & table)
{
	if (holder == NULL) {
		return "";
	}

	return MESSAGE(*holder).Compose(table);
}

//
// End-of-file ECG_MESSAGE.HPP
//
