
/*************************************************************************/
/* ECG_PTR.H                                                  29.07.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_PTR_H
#define ECG_PTR_H

namespace ecg {

/*************************************************************************/
/* class P                                                               */
/*                                                                       */
/*************************************************************************/
template <class T>
class P {
	template <class T0>
	friend class ecg::P;
public:
	P(void)
		: ptr(NULL), cntr(NULL)
	{}

	P(T * ptr)
		: ptr(ptr), cntr(NULL)
	{
		if (ptr) {
			cntr = new int(1);
		}
	}

	P(const P<T> & ptr)
		: ptr(ptr.ptr), cntr(ptr.cntr)
	{ addref(); }

	template <class T0>
	P(const P<T0> & ptr)
		: ptr(ptr.ptr), cntr(ptr.cntr)
	{ addref(); }

	~P(void)
	{
		release();

		// postcondition
		assert(ptr == NULL && cntr == NULL);
	}

public:
	T *				operator ->		(void) { assert(ptr); return ptr; }
	const T *		operator ->		(void) const { assert(ptr); return ptr; }
	
	T &				operator *		(void) { assert(ptr); return *ptr; }
	const T &		operator *		(void) const { assert(ptr); return *ptr; }

	bool			operator !		(void) const { return counter() == 0; }

	bool			operator ==		(const T * p) const { return ptr == p; }
	bool			operator !=		(const T * p) const { return ptr != p; }

	bool			operator ==		(const P<T> & p) const { assert(ptr != p.ptr || cntr == p.cntr); return ptr == p.ptr; }
	bool			operator !=		(const P<T> & p) const { assert(ptr != p.ptr || cntr == p.cntr); return ptr != p.ptr; }

	// why not?
	bool			operator <		(const P<T> & p) const { return *ptr < *p.ptr; }
	bool			operator <=		(const P<T> & p) const { return *ptr <= *p.ptr; }
	bool			operator >		(const P<T> & p) const { return *ptr > *p.ptr; }
	bool			operator >=		(const P<T> & p) const { return *ptr >= *p.ptr; }

	P<T> &			operator =		(T * p)
	{
		reset(p);
		return *this;
	}

	P<T> &			operator =		(const P<T> & p)
	{
		if (ptr == p.ptr) {
			return *this;
		}

		release();

		cntr = p.cntr;
		ptr = p.ptr;

		addref();

		return *this;
	}

	int 			counter			(void) const
	{
		if (cntr) {
			return *cntr;
		}

		return 0;
	}

	T *				get				(void) { return ptr; }
	const T *		get				(void) const { return ptr; }
	
	void			reset			(T * p = NULL)
	{
		if (ptr == p) {
			assert(ptr == NULL);
			return ;
		}

		release();

		if (p != NULL) {
			cntr = new int(1);
			ptr	= p;
		}
	}

private:
	void			addref			(void) const
	{
		if (ptr) {
			assert(cntr);
			++(*cntr);
		}
	}
	
	void			release			(void) const
	{
		if (counter() == 0) {
			assert(ptr == NULL && cntr == NULL);
			return ;
		}

		if (counter() == 1) {
			delete ptr;
			ptr = NULL;
			delete cntr;
			cntr = NULL;
			return ;
		}

		--(*cntr);
		ptr		= NULL;
		cntr	= NULL;
	}

private:
	// mutable because P<T> is out of const semantics because you can have
	// single const P<T> to const T * and it must be deleted anyway
	// irrelative to constancy
	mutable int *	cntr;
	mutable T *		ptr;
};

} // namespace ecg

#endif // ECG_PTR_H
