
/*************************************************************************/
/* ECG_STACK.H                                                29.07.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_STACK_H
#define ECG_STACK_H

#ifndef ECG_PTR_H
#error ecg_ptr.h must be included first
#endif

namespace ecg {

/*************************************************************************/
/* class PTR_STACK                                                       */
/*                                                                       */
/*************************************************************************/
class PTR_STACK {
	typedef P<void *> P_VOID;

	struct BACKUP {
		virtual void Destroy(P_VOID & ptr) = 0;
		virtual ~BACKUP(void) {}
	};

	template <class T>
	struct BACKUP_IMPL : BACKUP {
		virtual void Destroy(P_VOID & ptr)
		{ reinterpret_cast<P<T> &> (ptr).~P(); }
	};

	struct DATA {
		char ptr[sizeof(P_VOID)];
		char backup[sizeof(BACKUP)];
	};

public:
	~PTR_STACK		(void)
	{ clear(); }

	template<class T>
	void			push		(T * val)
	{ push(P<T> (val)); }

	template<class T>
	void			push		(P<T> val)
	{
		c.push(DATA());

		* reinterpret_cast<P<T> *> (c.top().ptr) = val;
		new (c.top().backup) BACKUP_IMPL<T> ();
	}

	void			pop			(void)
	{
		assert(!empty());
		reinterpret_cast<BACKUP *> (c.top().backup)->Destroy(
			* reinterpret_cast<P_VOID *> (c.top().ptr)
		);
		c.pop();
	}

	template<class T>
	P<T>			pop			(void)
	{
		assert(!empty());
		P<T> res = * reinterpret_cast<P<T> *> (c.top().ptr);
	
		assert(dynamic_cast<BACKUP_IMPL<T> *> (
			reinterpret_cast<BACKUP *> (c.top().backup)
		));

		c.pop();

		return res;
	}

	template<class T>
	P<T> &			top			(void) { return at<T> (0); }

	template<class T>
	const P<T> &	top			(void) const { return at<T> (0); }

	template<class T>
	P<T> &			at			(int index)
	{
		assert((size_t)index < c.size());

		DATA & data = c.c[size() - index - 1];

		assert(dynamic_cast<BACKUP_IMPL<T> *> (
			reinterpret_cast<BACKUP *> (data.backup)
		));

		return * reinterpret_cast<P<T> *> (data.ptr);
	}

	template<class T>
	const P<T> &	at			(int index) const
	{ return const_cast<STACK *> (this)->at(index); }

	bool			empty		(void) const { return c.empty(); }
	int				size		(void) const { return (int)c.size(); }

	void			clear		(void)
	{
		while (!empty()) {
			pop();
		}
	}

private:
	std::stack<DATA> c;
};

class RAW_STACK {
public:
	RAW_STACK(int item_size)
		: item_size(item_size), data(), length(), allocated()
	{}

	~RAW_STACK(void)
	{
		delete [] data;
	}

	template <class T>
	inline void		push			(const T & val)
	{
		assert(sizeof(T) <= item_size);

		if (length == allocated) {
			if (allocated == 0) {
				allocated = 1;
			}
			do {
				allocated <<= 1;
			} while (allocated <= length);
			
			char * prev = data;
			data = new char[item_size * allocated];

			if (prev != NULL) {
				memcpy(data, prev, item_size * length);
			}
		}

		* reinterpret_cast<T *> (data + item_size * length) = val;
		++length;
	}

	template <class T>
	inline T &		top				(int index)
	{
		assert(length - index > 0);
		assert(sizeof(T) <= item_size);
		return * reinterpret_cast<T *> (data + (length - 1 - index) * item_size);
	}

	template <class T>
	inline const T & top			(int index) const
	{
		assert(length > 0);
		return * reinterpret_cast<const T *> (data + (length - 1 - index) * item_size);
	}

	template <class T>
	inline T &		top				(void)
	{
		assert(length > 0);
		assert(sizeof(T) <= item_size);
		return * reinterpret_cast<T *> (data + (length - 1) * item_size);
	}

	template <class T>
	inline const T & top			(void) const
	{
		assert(length > 0);
		return * reinterpret_cast<const T *> (data + (length - 1) * item_size);
	}

	inline void		pop				(void)
	{
                assert(length > 0);
		--length;
	}

	inline int		size			(void) const { return length; }
	inline bool		empty			(void) const { return length == 0; }

private:
	const int		item_size;
	char *			data;
	int				length;
	int				allocated;
};

} // namespace ecg

#endif // ECG_STACK_H
