
/*************************************************************************/
/* ECG_ALLOCATOR.H                                            06.09.2007 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_ALLOCATOR_H
#define ECG_ALLOCATOR_H

#pragma once

namespace ecg {

template <class CHILD>
class CLASS_ALLOC_POLICY {
public:
	template <class T>
	inline T *			Alloc			(void);
	inline int			Destruct		(char * ptr);

private:
	struct HEADER {
		virtual int		Destruct		(void * obj) const = 0;
	};

	template <class T>
	struct HEADER_IMPL {
		virtual int		Destruct		(void * obj) const { reinterpret_cast<T *> (obj)->~T(); return sizeof(T); }
	};

	enum {
		HEADER_SIZE = sizeof(HEADER_IMPL<int>),
	};

protected:
	~CLASS_ALLOC_POLICY(void) {}
};

template <class CHILD>
class POD_ALLOC_POLICY {
public:
	template <class T>
	inline T *			Alloc			(void);
	inline char *		Alloc			(int size);
	inline int			Destruct		(char * ptr);

protected:
	~POD_ALLOC_POLICY(void) {}
};

template <class T>
class T_BUFFER_POLICY {
public:
	template <class CHILD>
	struct POLICY {
		T *				Alloc			(void);
		int				Destruct		(char * ptr);
	};
};

template <template <class> class ALLOC_POLICY, int BLOCK_SIZE>
class BUFFER_ALLOCATOR : public ALLOC_POLICY<BUFFER_ALLOCATOR<ALLOC_POLICY, BLOCK_SIZE> > {
	typedef ALLOC_POLICY<BUFFER_ALLOCATOR> POLICY_T;
	friend POLICY_T;
public:
	BUFFER_ALLOCATOR	(void) 
		: root(), rootFree(), rootFreeParent()
	{}

	~BUFFER_ALLOCATOR	(void) { Clear(); }

	void				Clear				(void);

private:
	BUFFER_ALLOCATOR(const BUFFER_ALLOCATOR &);
	void operator = (const BUFFER_ALLOCATOR &);

	char *				intAlloc			(int size);

private:
	struct BLOCK {
		BLOCK *			next;
		char *			bufFree;
		char			buf[1];
	};

	BLOCK *				root;
	BLOCK *				rootFree;
	BLOCK **			rootFreeParent;
};

typedef BUFFER_ALLOCATOR<CLASS_ALLOC_POLICY, 4096> CLASS_BUFFER;
typedef BUFFER_ALLOCATOR<POD_ALLOC_POLICY,   4096> POD_BUFFER;

template <class T>
class T_BUFFER : public BUFFER_ALLOCATOR<T_BUFFER_POLICY<T>::POLICY, 4096> {};

#include "ecg_allocator.hpp"

} // namespace ecg

#endif // ECG_ALLOCATOR_H