
/************************************************************************/
/* ecg::CLASS_ALLOC_POLICY<CHILD, T>::Alloc
/*
/************************************************************************/
template <class CHILD> template <class T>
T *	CLASS_ALLOC_POLICY<CHILD>::Alloc(void)
{
	char * res = static_cast<CHILD *> (this)->intAlloc(HEADER_SIZE + sizeof(T));
	new (res) HEADER_IMPL<T> ();
	return new (res + HEADER_SIZE) T();
}

/************************************************************************/
/* ecg::CLASS_ALLOC_POLICY<CHILD, T>::Destruct
/*
/************************************************************************/
template <class CHILD>
int CLASS_ALLOC_POLICY<CHILD>::Destruct(char * ptr)
{
	HEADER * header = reinterpret_cast<HEADER *> (ptr);
	return HEADER_SIZE + header->Destruct(ptr + HEADER_SIZE);
}

/************************************************************************/
/* ecg::BUFFER::Alloc
/*
/************************************************************************/
template <template <class> class ALLOC_POLICY, int BLOCK_SIZE>
char * BUFFER_ALLOCATOR<ALLOC_POLICY, BLOCK_SIZE>::intAlloc(int size)
{
	assert(size + sizeof(BLOCK) <= BLOCK_SIZE);

	for (;;) {
		if (rootFree == NULL) {
			if (rootFreeParent == NULL) {
				rootFree = root = (BLOCK *) new char[BLOCK_SIZE];;
			} else {
				rootFree = *rootFreeParent = (BLOCK *) new char[BLOCK_SIZE];;
			}

			rootFree->next		= NULL;
			rootFreeParent		= &rootFree->next;
			rootFree->bufFree	= rootFree->buf + sizeof(BLOCK);
		}

		int remains = BLOCK_SIZE - static_cast<int> (rootFree->bufFree - (char *)rootFree);
		if (remains >= size) {
			char * res = rootFree->bufFree;
			rootFree->bufFree += size;

			return res;
		}

		rootFree = rootFree->next;
	}
}

/************************************************************************/
/* ecg::BUFFER::Free
/*
/************************************************************************/
template <template <class> class ALLOC_POLICY, int BLOCK_SIZE>
void BUFFER_ALLOCATOR<ALLOC_POLICY, BLOCK_SIZE>::Clear(void)
{
	BLOCK * cur = root;
	while (cur != NULL) {
		char * ptr = cur->buf + sizeof(BLOCK);
		while (ptr != cur->bufFree) {
			ptr += POLICY_T::Destruct(ptr);
		}

		char * block = reinterpret_cast<char *> (cur);
		cur = cur->next;

		delete [] block;
	}

	root = rootFree = NULL;
	rootFreeParent = NULL;
}

/************************************************************************/
/* ecg::POD_ALLOC_POLICY<CHILD, T>::Alloc
/*
/************************************************************************/
template <class CHILD> template <class T>
T * POD_ALLOC_POLICY<CHILD>::Alloc(void)
{
	return (T *) Alloc(sizeof(T));
}

/************************************************************************/
/* ecg::POD_ALLOC_POLICY::Alloc
/*
/************************************************************************/
template <class CHILD>
char * POD_ALLOC_POLICY<CHILD>::Alloc(int size)
{
	char * res = static_cast<CHILD *> (this)->intAlloc(size);
	*(int *)res = size;
	return res + sizeof(int);
}

/************************************************************************/
/* ecg::POD_ALLOC_POLICY<CHILD>::Destruct
/*
/************************************************************************/
template <class CHILD>
int POD_ALLOC_POLICY<CHILD>::Destruct(char * ptr)
{
	return *(int *)ptr;
}

/************************************************************************/
/* ecg::T_BUFFER_HELPER::POLICY<T, CHILD>::Alloc
/*
/************************************************************************/
template <class T> template <class CHILD>
T * T_BUFFER_POLICY<T>::POLICY<CHILD>::Alloc(void)
{
	char * res = static_cast<CHILD *> (this)->intAlloc(sizeof(T));
	return new (res) T;
}

/************************************************************************/
/* ecg::T_BUFFER_HELPER::POLICY<T, CHILD>::Destruct
/*
/************************************************************************/
template <class T> template <class CHILD>
int T_BUFFER_POLICY<T>::POLICY<CHILD>::Destruct(char * ptr)
{
	reinterpret_cast<T *> (ptr)->~T();
	return sizeof(T);
}