
#ifndef ECG_STR_POOL_H
#define ECG_STR_POOL_H

#pragma once

namespace ecg {

class STR_ID;
class STR_POOL;

class STR_ID {
	friend class STR_POOL;
public:
	inline const char *	c_str			(void) const;
	inline const std::string & str		(void) const;

	// operator < is used for search
	inline bool			operator <		(STR_ID p) const { return str() <  p.str(); }
	inline bool			operator ==		(STR_ID p) const { return id == p.id; }
	inline bool			operator !=		(STR_ID p) const { return id != p.id; }

	inline bool			isValid			(void) const { return id != -1; }

private:
	int					id;
};

class STR_POOL {
	friend class STR_ID;
public:
	static STR_POOL *	Get				(void) { static STR_POOL pool; return &pool; }

	inline STR_ID		Add				(const std::string & str);

	inline void			Clear			(void);

private:
	STR_POOL			(void) {}

private:
	std::vector<std::string> pool;
	std::set<STR_ID>	poolId;
};

inline STR_ID toSTR_ID(const std::string & str) { return STR_POOL::Get()->Add(str); }

#include "ecg_str_pool.hpp"

};

#endif // ECG_STR_POOL_H