
/*************************************************************************/
/* ECG_GEN_COMPILER.CPP                                       28.07.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "ecg_generator.h"

using namespace std;
using namespace ecg;

/*************************************************************************/
/* GENERATOR::PrintActionTableItem()                                     */
/*                                                                       */
/*************************************************************************/
void GENERATOR::PrintActionTableItem(const LALR_TABLE_LINE & line, RULE::ITEM item, RULE_NUMBER_MAP & ruleNumber, ostream & code)
{
	// look up action
	LALR_TABLE_LINE::const_iterator action = line.find(item);

	// check for error
	if (action == line.end()) {
		if (item.isTerm) {
			code << "{}, ";
		} else {
			code << "-1, ";
		}
		return ;
	}

	// check for shift
	if (action->second.isShift) {
		if (item.isTerm) {
			code << "{ AUTO_T::SHIFT, " << action->second.state << " }, ";
		} else {
			code << action->second.state << ", ";
		}
		return ;
	}

	assert(item.isTerm);
	assert(ruleNumber.find(action->second.furl) != ruleNumber.end());
	code << "{ AUTO_T::REDUCE, " << ruleNumber[action->second.furl] << "}, ";
}

/*************************************************************************/
/* GENERATOR::PrepareLiteral()                                           */
/*                                                                       */
/*************************************************************************/
std::string GENERATOR::ConvertToLiteral(const std::string & value)
{
	std::string res = "\"";
	for (int i=0; i<(int)value.size(); ++i) {
		switch (value[i]) {
			case '\\': res += "\\\\"; break;
			case '\"': res += "\\\""; break;
			case '\n': res += "\\n";  break;
			default:   res += value[i]; break;
		}
	}
	return res + "\"";
}

/*************************************************************************/
/* GENERATOR::GlobalTypeName()                                           */
/*                                                                       */
/*************************************************************************/
std::string GENERATOR::GlobalTypeName(const std::string & typeName)
{
	if (strncmp(typeName.c_str(), "IMPL_T", 6) == 0) {
		return name + "::" + typeName;
	}
	return typeName;
}

/*************************************************************************/
/* GENERATOR::Generate()                                                 */
/*                                                                       */
/*************************************************************************/
void GENERATOR::Generate(std::ostream & header, std::ostream & code, std::ostream & hlp)
{
	string nameUpr = name;
	assert(strlen(&nameUpr[0]) == nameUpr.size());
	strupr(&nameUpr[0]);

	STRING_ARRAY nonterminalAlias(nonterminal.size());
	for (size_t i=0; i<nonterminalAlias.size(); ++i) {
		nonterminalAlias[i] = nontermPrefix + TransformNameToAlias(nonterminal[i].name);
	}

	// <<<  >>>

	header << endl;
	header << "/*************************************************************************/" << endl;
	header << "/* " << nameUpr << ".H" << string(max(0, 64 - (int)nameUpr.size()), ' ') << "E13 */" << endl;
	header << "/*                                                                       */" << endl;
	header << "/* .                                                                     */" << endl;
	header << "/*************************************************************************/" << endl;
	header << endl;
	header << "#ifndef " << nameUpr << "_H" << endl;
	header << "#define " << nameUpr << "_H" << endl;
	header << endl;
	header << "#ifndef TOKEN_EOF" << endl;
	header << "#define TOKEN_EOF		-1" << endl;
	header << "#endif" << endl;
	header << endl;
	header << "#ifndef TOKEN_ERROR" << endl;
	header << "#define TOKEN_ERROR		-2" << endl;
	header << "#endif" << endl;
	header << endl;
	header << "namespace ecg {" << endl;
	header << endl;
	header << "class " << nameImpl << ";" << endl;
	header << endl;

	// for each rule calc its number
	map<RULE_INDEX, int> ruleNumber;
	{
		int ruleCount = 0;
		RULE_INDEX index;
		for (NONTERMINAL_ARRAY::const_iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
			index.nontermId		= nonterm->id;
			index.ruleNumber	= 0;
			for (RULE_ARRAY::const_iterator rule = nonterm->rules.begin(); rule != nonterm->rules.end(); ++rule) {
				assert(ruleNumber.find(index) == ruleNumber.end());
				ruleNumber[index] = ruleCount;

				++ruleCount;
				++index.ruleNumber;
			}
		}
		assert(ruleCount == ruleNumber.size());
	}

	header << "/*************************************************************************/" << endl;
	header << "/* class " << name << string(max(0, 64 - (int)name.size()), ' ') << "*/" << endl;
	header << "/*                                                                       */" << endl;
	header << "/*************************************************************************/" << endl;
	header << "class " << name << " {" << endl;
	header << "public:" << endl;
	header << "	enum {" << endl;
	header << "		STATE_COUNT		= " << (int)lalrTable.size() << "," << endl;
	header << "		TERM_COUNT		= " << (int)terminal.size() << "," << endl;
	header << "		NONTERM_COUNT	= " << (int)nonterminal.size() << "," << endl;
	header << "		RULE_COUNT		= " << (int)ruleNumber.size() << "," << endl;
	header << "	};" << endl;
	header << endl;
	header << "	//enum TERMINALS {" << endl;

	int maxlen = 0;
	for (TERMINAL_ARRAY::const_iterator term = terminal.begin(); term != terminal.end(); ++term) {
		maxlen = max(maxlen, (int)term->alias.size());
	}
	for (STRING_ARRAY::const_iterator nonterm = nonterminalAlias.begin(); nonterm != nonterminalAlias.end(); ++nonterm) {
		maxlen = max(maxlen, (int)nonterm->size());
	}
	maxlen = 1 + (maxlen + 1 + 3) / 4;

	for (TERMINAL_ARRAY::const_iterator term = terminal.begin(); term != terminal.end(); ++term) {
		const string & alias = terminal[term->id].alias;

		header << "	//		" << alias << ",";
		int count = ((int)alias.size() + 2 + 3) / 4;
		for (int i=0; i<maxlen - count; ++i) {
			header << "\t";
		}
		assert(term->name.size() > 0);
		if (term->name[0] == '$') {
			header << "	// " << term->name << endl;
		} else {
			header << "	// \"" << term->name << "\"" << endl;
		}
	}

	header << "	//};" << endl;
	header << endl;
	header << "	enum NONTERMINALS {" << endl;

	for (NONTERMINAL_ARRAY::const_iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
		assert((size_t)nonterm->id < nonterminalAlias.size());
		const string & alias = nonterminalAlias[nonterm->id];

		header << "		" << alias << ",";
		int count = ((int)alias.size() + 2 + 3) / 4;
		for (int i=0; i<maxlen - count; ++i) {
			header << "\t";
		}
		header << "// " << nonterm->name << endl;
	}

	header << "	};" << endl;
	header << endl;

	// collect list of value types
	STRING_STRING_MAP valueTypes;
	{
		int valueTypesCounter = 0;
		for (AUTOFUNCTION_MAP::const_iterator af = autoFuncs.begin(); af != autoFuncs.end(); ++af) {
			// params
			for (int i=0; i<(int)af->second->params.size(); ++i) {
				const AUTO_FUNCTION::PARAM & prm = af->second->params[i];
				if (prm.isValue && prm.name != ecg::VOID_TYPE && prm.name != ecg::STRING_TYPE) {
					if (valueTypes.find(prm.name) == valueTypes.end()) {
						valueTypes.insert(make_pair(prm.name, "type_" + string(1, 'a' + valueTypesCounter++)));
					}
				}
			}

			// result
			const AUTO_FUNCTION::PARAM & prm = af->second->result;
			if (prm.isValue && prm.name != ecg::VOID_TYPE && prm.name != ecg::STRING_TYPE) {
				if (valueTypes.find(prm.name) == valueTypes.end()) {
					valueTypes.insert(make_pair(prm.name, "type_" + string(1, 'a' + valueTypesCounter++)));
				}
			}
		}
	}

	if (valueTypes.empty()) {
		header << "	typedef int VALUE_T;" << endl;
	} else {
		header << "	union VALUE_T;" << endl;
	}
	header << "	typedef " << nameImpl << " IMPL_T;" << endl;
	header << endl;
	header << "public:" << endl;

	maxlen = max(0, (int)(16 - name.size() + 3) / 4);

	header << "						" << name << (maxlen == 0 ? " " : string(maxlen, '\t')) << "(void);" << endl;
	header << "	void				Compile				(void);" << endl;
	header << endl;

	header << "protected:" << endl;
	header << "	void				intPushString		(void);" << endl;
	header << "	void				intPushEmptyString	(void);" << endl;
	header << "	void				intCreateStringArray(void);" << endl;
	header << "	void				intAppendStringArray(void);" << endl;
	header << endl;
	header << "	template <class T> void intCreatePtrArray(void);" << endl;
	header << "	template <class T> void intAppendPtrArray(void);" << endl;
	header << "	template <class T> void intCreateValArray(void);" << endl;
	header << "	template <class T> void intAppendValArray(void);" << endl;
	header << endl;

	if (!autoFuncs.empty()) {
		header << "protected:" << endl;
		for (AUTOFUNCTION_MAP::const_iterator af = autoFuncs.begin(); af != autoFuncs.end(); ++af) {
			header << "	void				" << af->first << "		(void);" << endl;
		}
		header << endl;
	}

	header << "protected:" << endl;
	header << "	COMPILER_STATE_F	compilerState;" << endl;
	header << "	PTR_STACK			pstack;" << endl;
	header << "	std::stack<std::string> sstack;" << endl;
	header << "	RAW_STACK			vstack;" << endl;
	header << endl;
	header << "private:" << endl;
	header << "	IMPL_T &			Impl			(void);" << endl;
	header << endl;
	header << "private:" << endl;
	header << "	typedef AUTO_CONFIG<STATE_COUNT, TERM_COUNT, NONTERM_COUNT, RULE_COUNT> CONFIG_T;" << endl;;
	header << "	typedef AUTO<IMPL_T, CONFIG_T> AUTO_T;" << endl;
	header << "	friend AUTO_T;" << endl;
	header << endl;
	header << "	typedef AUTO_T::ACTION		ACTION;" << endl;
	header << "	typedef AUTO_T::ACTION_PROC	ACTION_PROC;" << endl;
	header << "	typedef AUTO_T::PRODUCTION	PRODUCTION;" << endl;
	header << endl;
	header << "	static const ACTION			actionTable			[STATE_COUNT][TERM_COUNT];" << endl;
	header << "	static const ACTION			actionTableEof		[STATE_COUNT];" << endl;
	header << "	static const ACTION			actionTableError	[STATE_COUNT];" << endl;
	header << "	static const int			productionTable		[STATE_COUNT][NONTERM_COUNT];" << endl;
	header << "	static const PRODUCTION		procTableProduction	[RULE_COUNT];" << endl;
	header << "	static const ACTION_PROC	procTableTerm		[TERM_COUNT];" << endl;
	header << "	static const ACTION_PROC	procTableNonterm	[NONTERM_COUNT];" << endl;
	header << endl;

	header << "	AUTO_T				compilerAuto;" << endl;
	header << "};" << endl;
	header << endl;
	header << "} // namespace ecg" << endl;
	header << endl;
	header << "#endif // " << nameUpr << "_H" << endl;
	header << endl;

	// <<<   >>>

	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* " << nameUpr << ".CPP" << string(max(0, 62 - (int)nameUpr.size()), ' ') << "E13 */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/* .                                                                     */" << endl;
	code << "/*************************************************************************/" << endl;
	code << endl;
	code << "#include \"stdafx.h\"" << endl;
	code << "#include \"" << implName << ".h\"" << endl;
	code << endl;
	code << "using namespace ecg;" << endl;
	code << "using namespace std;" << endl;
	code << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< ACTION TABLE >>>
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::actionTable" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const " << name << "::ACTION " << name << "::actionTable[STATE_COUNT][TERM_COUNT] = {" << endl;
	// for each table line
	for (LALR_TABLE::const_iterator line = lalrTable.begin(); line != lalrTable.end(); ++line) {
		code << "	{ ";
		// for each terminal
		for (TERMINAL_ARRAY::const_iterator term = terminal.begin(); term != terminal.end(); ++term) {
			PrintActionTableItem(*line, RULE::ITEM(true, term->id), ruleNumber, code);
		}
		code << "}," << endl;
	}
	code << "};" << endl;
	code << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< ACTION TABLE FOR $eof TERMINAL >>>
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::actionTableEof" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const " << name << "::ACTION " << name << "::actionTableEof[STATE_COUNT] = {" << endl;
	code << "	";
	// for each table line
	for (LALR_TABLE::const_iterator line = lalrTable.begin(); line != lalrTable.end(); ++line) {
		PrintActionTableItem(*line, RULE::ITEM(true, TOKEN_EOF), ruleNumber, code);
	}
	code << "};" << endl;
	code << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< ACTION TABLE FOR $error TERMINAL >>>
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::actionTableError" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const " << name << "::ACTION " << name << "::actionTableError[STATE_COUNT] = {" << endl;
	code << "	";
	// for each table line
	for (LALR_TABLE::const_iterator line = lalrTable.begin(); line != lalrTable.end(); ++line) {
		PrintActionTableItem(*line, RULE::ITEM(true, TOKEN_ERROR), ruleNumber, code);
	}
	code << "};" << endl;
	code << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< PRODUCTION TABLE >>>
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::productionTable" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const int " << name << "::productionTable[STATE_COUNT][NONTERM_COUNT] = {" << endl;
	// for each table line
	for (LALR_TABLE::const_iterator line = lalrTable.begin(); line != lalrTable.end(); ++line) {
		code << "	{ ";
		// for each nonterminal
		for (NONTERMINAL_ARRAY::const_iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
			PrintActionTableItem(*line, RULE::ITEM(false, nonterm->id), ruleNumber, code);
		}
		code << "}," << endl;
	}
	code << "};" << endl;
	code << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< PROCEDURE TABLE PRODUCTION >>>
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::procTableProduction" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const " << name << "::PRODUCTION " << name << "::procTableProduction[RULE_COUNT] = {" << endl;
	int counter = 0;
	for (NONTERMINAL_ARRAY::const_iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
		RULE_INDEX index;
		index.nontermId		= nonterm->id;
		index.ruleNumber	= 0;

		for (RULE_ARRAY::const_iterator rule = nonterm->rules.begin(); rule != nonterm->rules.end(); ++rule) {
			code << "	{ ";
			assert(ruleNumber[index] == counter);
			if (!rule->code.empty()) {
				code << "&" << nameImpl << "::" << rule->code << ", ";
			} else {
				code << "NULL, ";
			}
			code << (int)rule->items.size() << ", " << nonterm->id << " }," << endl;

			++counter;
			++index.ruleNumber;
		}
	}
	code << "};" << endl;
	code << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< PROCEDURE TABLE TERMINAL >>>
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::procTableTerm" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const " << name << "::ACTION_PROC " << name << "::procTableTerm[TERM_COUNT] = {" << endl;
	for (TERMINAL_ARRAY::const_iterator term = terminal.begin(); term != terminal.end(); ++term) {
		if (!term->code.empty()) {
			code << "	&" << nameImpl << "::" << term->code << "," << endl;
		} else {
			code << "	NULL," << endl;
		}
	}
	code << "};" << endl;
	code << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< PROCEDURE TABLE NONTERMINAL >>>
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::procTableNonterm" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const " << name << "::ACTION_PROC " << name << "::procTableNonterm[NONTERM_COUNT] = {" << endl;
	for (NONTERMINAL_ARRAY::const_iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
		if (!nonterm->code.empty()) {
			code << "	&" << nameImpl << "::" << nonterm->code << "," << endl;
		} else {
			code << "	NULL," << endl;
		}
	}
	code << "};" << endl;
	code << endl;

	if (!valueTypes.empty()) {
		code << "/*************************************************************************/" << endl;
		code << "/* " << name << "::VALUE_T" << string(max(0, int(77 - 2 * name.size())), ' ') << "*/" << endl;
		code << "/*                                                                       */" << endl;
		code << "/*************************************************************************/" << endl;
		code << "union " << name << "::VALUE_T {" << endl;
		for (STRING_STRING_MAP::const_iterator vt = valueTypes.begin(); vt != valueTypes.end(); ++vt) {
			code << "	" << GlobalTypeName(vt->first)
				<< "	" << vt->second << ";" << endl;
		}
		code << "};" << endl;
		code << endl;
	}

	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::" << name << "()" << string(max(0, int(66 - 2 * name.size())), ' ') << "*/" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << name << "::" << name << "(void)" << endl;
	code << "	: vstack(sizeof(VALUE_T))" << endl;
	code << "{" << endl;
	code << "	compilerAuto.actionTable		= &actionTable;" << endl;
	code << "	compilerAuto.actionTableEof		= actionTableEof;" << endl;
	code << "	compilerAuto.actionTableError	= actionTableError;" << endl;
	code << "	compilerAuto.productionTable	= &productionTable;" << endl;
	code << "	compilerAuto.procTableProduction = procTableProduction;" << endl;
	code << "	compilerAuto.procTableTerm		= procTableTerm;" << endl;
	code << "	compilerAuto.procTableNonterm	= procTableNonterm;" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* Impl()                                                                */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << nameImpl << " & " << name << "::Impl(void)" << endl;
	code << "{" << endl;
	code << "	return * static_cast<IMPL_T *> (this);" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* intPushString()                                                       */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "void " << name << "::intPushString(void)" << endl;
	code << "{" << endl;
	code << "	sstack.push(Impl().TokenValue());" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* intPushEmptyString()                                                  */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "void " << name << "::intPushEmptyString(void)" << endl;
	code << "{" << endl;
	code << "	sstack.push(std::string());" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* intCreateStringArray()                                                */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "void " << name << "::intCreateStringArray(void)" << endl;
	code << "{" << endl;
	code << "	pstack.push(new std::vector<std::string>());" << endl;
	code << "	pstack.top<std::vector<std::string> >()->push_back(sstack.top());" << endl;
	code << "	sstack.pop();" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* intAppendStringArray()                                                */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "void " << name << "::intAppendStringArray(void)" << endl;
	code << "{" << endl;
	code << "	pstack.top<std::vector<std::string> >()->push_back(sstack.top());" << endl;
	code << "	sstack.pop();" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* intCreatePtrArray()                                                   */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "template <class T> void " << name << "::intCreatePtrArray(void)" << endl;
	code << "{" << endl;
	code << "	P<T> a = pstack.pop<T>();" << endl;
	code << "	pstack.push(new std::vector<P<T> >());" << endl;
	code << "	pstack.top<std::vector<P<T> > >()->push_back(a);" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* intAppendPtrArray()                                                   */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "template <class T> void " << name << "::intAppendPtrArray(void)" << endl;
	code << "{" << endl;
	code << "	P<T> a = pstack.pop<T>();" << endl;
	code << "	pstack.top<std::vector<P<T> > >()->push_back(a);" << endl;
	code << "}" << endl;
	code << endl;

	// generateo value type helper class
	code << "template <class T>" << endl;
	code << "struct VALUE_TYPE_HELPER;" << endl;
	code << endl;

	for (STRING_STRING_MAP::const_iterator vt = valueTypes.begin(); vt != valueTypes.end(); ++vt) {
		code << "template <>" << endl;
		code << "struct VALUE_TYPE_HELPER<" << GlobalTypeName(vt->first) << " > {" << endl;
		code << "	static inline const " << GlobalTypeName(vt->first)
			 << " & get(const " << name << "::VALUE_T & p) { return p." << vt->second << "; }" << endl;
		code << "};" << endl;
		code << endl;
	}

	code << "/*************************************************************************/" << endl;
	code << "/* intCreateValArray()                                                   */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "template <class T> void " << name << "::intCreateValArray(void)" << endl;
	code << "{" << endl;
	code << "	T a = VALUE_TYPE_HELPER<T>::get(vstack.top<VALUE_T>());" << endl;
	code << "	vstack.pop();" << endl;
	code << "	pstack.push(new std::vector<T>());" << endl;
	code << "	pstack.top<std::vector<T> >()->push_back(a);" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* intAppendValArray()                                                   */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "template <class T> void " << name << "::intAppendValArray(void)" << endl;
	code << "{" << endl;
	code << "	T a = VALUE_TYPE_HELPER<T>::get(vstack.top<VALUE_T>());" << endl;
	code << "	vstack.pop();" << endl;
	code << "	pstack.top<std::vector<T> >()->push_back(a);" << endl;
	code << "}" << endl;
	code  << endl;

	for (AUTOFUNCTION_MAP::const_iterator af = autoFuncs.begin(); af != autoFuncs.end(); ++af) {
		code << "/*************************************************************************/" << endl;
		code << "/* " << name << "::" << af->first << "()" << string(max(0, int(66 - name.size() - af->first.size())), ' ') << "*/" << endl;
		code << "/*                                                                       */" << endl;
		code << "/*************************************************************************/" << endl;

		code << "void " << name << "::" << af->first << "(void)" << endl;
		code << "{" << endl;

		const AUTO_FUNCTION & func = *af->second;

		// production params
		AUTO_FUNCTION::PARAM type;
		if (!func.isMarker) {
			for (int ri = (int)func.params.size() - 1; ri >= 0; --ri) {
				type = func.params[ri];

				if (type.name == STRING_TYPE) {
					assert(type.isValue);

					if (!func.isConstructor || func.result.name != VOID_TYPE) {
						code << "	std::string " << char('a' + ri) << " = sstack.top();" << endl;
					}
					code << "	sstack.pop();" << endl;
					continue;
				}
				
				if (!func.isConstructor || func.result.name != VOID_TYPE) {
					if (type.isValue) {
						assert(valueTypes.find(type.name) != valueTypes.end());
						code << "	" << type.name << " " << char('a' + ri) << " = vstack.top<VALUE_T>()." << valueTypes[type.name] << ";" << endl;
						code << "	vstack.pop();" << endl;
					} else {
						code << "	P<" << type.name << " > " << char('a' + ri) << " = pstack.pop<" << type.name << " >();" << endl;
					}
				} else {
					if (type.isValue) {
						code << "	vstack.pop();" << endl;
					} else {
						code << "	pstack.pop();" << endl;
					}
				}
			}
		} else {
			assert(!func.isConstructor || func.result.name != VOID_TYPE && !func.result.isValue);

			// don't remove elements from stack when marker call
			int stringCounter = 0;
			int ptrCounter = 0;
			int valCounter = 0;
			for (int ri = (int)func.params.size() - 1; ri >= 0; --ri) {
				type = func.params[ri];
				if (type.name == STRING_TYPE) {
					assert(type.isValue);
					int index = ++stringCounter;
					code << "	assert(sstack.size() >= " << index << ");" << endl;
					code << "	std::string & " << char('a' + ri) << " = sstack.c[sstack.size() - " << index << "];" << endl;
					continue;
				}

				if (type.isValue) {
					int index = valCounter++;
					assert(valueTypes.find(type.name) != valueTypes.end());
					code << "	" << type.name << " & " << char('a' + ri) << " = vstack.top<VALUE_T>(" << index << ")." << valueTypes[type.name] << ";" << endl;
				} else {
					int index = ptrCounter++;
					code << "	P<" << type.name << " > & " << char('a' + ri) << " = pstack.at<" << type.name << " >(" << index << ");" << endl;
				}
			}
		}

		// result
		if (func.result.name != VOID_TYPE) {
			if (func.result.name == STRING_TYPE) {
				assert(func.result.isValue);
				assert(!func.isConstructor);
				code << "	std::string result;" << endl;
			} else {
				if (func.result.isValue) {
					assert(!func.isConstructor);
					code << "	" << func.result.name << " result;" << endl;
				} else {
					if (func.isConstructor) {
						code << "	pstack.push(new " << func.result.name << "(";
					} else {
						code << "	P<" << func.result.name << " > result;" << endl;
					}
				}
			}
		}

		if (!func.isConstructor) {
			// call
			code << "	Impl()." << func.name << "(";
		}

		// const params
		for (STRING_ARRAY::const_iterator param = func.constParams.begin(); param < func.constParams.end(); ++param) {
			code << *param;
			if (param + 1 != func.constParams.end()) {
				code << ", ";
			}
		}

		if (!func.isConstructor || func.result.name != VOID_TYPE) {
			// production params
			if (!func.constParams.empty() && (!func.params.empty() || !func.isConstructor && func.result.name != VOID_TYPE)) {
				code << ", ";
			}
			for (int i=0; (size_t)i < func.params.size(); ++i) {
				code << char('a' + i);
				if (i + 1 < (int)func.params.size()) {
					code << ", ";
				}
			}
		}

		if (!func.isConstructor) {
			// result
			if (func.result.name != VOID_TYPE) {
				if (!func.params.empty()) {
					code << ", ";
				}
				code << "result";
			}
		}

		if (!func.isConstructor) {
			code << ");" << endl;
			if (func.result.name != VOID_TYPE) {
				if (func.result.name != STRING_TYPE) {
					if (func.result.isValue) {
						code << "	vstack.push(VALUE_T());" << endl;
						assert(valueTypes.find(func.result.name) != valueTypes.end());
						code << "	vstack.top<VALUE_T>()." << valueTypes[func.result.name] << " = result;" << endl;
					} else {
						code << "	assert(Impl().compilerState.is(CST_ERROR_MASK) || result != NULL);" << endl;
						code << "	pstack.push(result);" << endl;
					}
				} else {
					assert(func.result.isValue);
					code << "	sstack.push(result);" << endl;
				}
			}
		} else {
			if (func.result.name != VOID_TYPE) {
				code << "));" << endl;
			}
		}

		code << "}" << endl;
		code << endl;
	}

	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::Compile()"<< string(max(0, 59 - (int)name.size()), ' ') << "*/" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "void " << name << "::Compile(void)" << endl;
	code << "{" << endl;
	code << "	compilerAuto.Compile(static_cast<IMPL_T &> (*this));" << endl;
	code << "}" << endl;
	
	// <<<     >>>

	hlp << "<<< GRAMMAR >>>" << endl;
	PrintGrammatics(hlp);
	hlp << "<<< STATES >>>" << endl;
	PrintLalrStates(hlp);

#if 0
	// don't print table - it's very huge, but not enough informative (for this size :))
	hlp << "<<< COMPILER TABLE >>>" << endl;
	PrintLalrTable(hlp);
#endif
}

/*************************************************************************/
/* GENERATOR::GenerateMsg()                                              */
/*                                                                       */
/*************************************************************************/
void GENERATOR::GenerateMsg(std::ostream & header, std::ostream & code)
{
	string name = GetName() + "_MSG";

	string nameUpr = name;
	assert(strlen(&nameUpr[0]) == nameUpr.size());
	strupr(&nameUpr[0]);

	header << endl;
	header << "/*************************************************************************/" << endl;
	header << "/* MSG                                                               E13 */" << endl;
	header << "/*                                                                       */" << endl;
	header << "/* .                                                                     */" << endl;
	header << "/*************************************************************************/" << endl;
	header << endl;
	header << "#ifndef " << nameUpr << "_H" << endl;
	header << "#define " << nameUpr << "_H" << endl;
	header << endl;
	header << "namespace ecg {" << endl;
	header << endl;
	assert(!messages.empty());
	header << "enum " << GetName() << "_MSG {" << endl;
	for (MESSAGE_MAP::const_iterator msg = messages.begin(); msg != messages.end(); ++msg) {
		header << "	MESSAGE_" << msg->first << "," << endl;
	}
	header << "};" << endl;
	header << endl;
	header << "extern const MSG_HOLDER * " << GetName() << "_GetMessage(const std::string & id);" << endl;
	header << "extern const MSG_HOLDER * " << GetName() << "_GetMessageByIndex(int index);" << endl;
	header << endl;
	header << "extern const MSG_HOLDER " << GetName() << "_messageTable[];" << endl;
	header << endl;
	header << "} // namespace ecg" << endl;
	header << endl;
	header << "#endif // " << nameUpr << "_H" << endl;
	header << endl;

	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* MSG                                                               E13 */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/* .                                                                     */" << endl;
	code << "/*************************************************************************/" << endl;
	code << endl;
	code << "#include \"stdafx.h\"" << endl;
	code << "#include \"" << implName << ".h\"" << endl;
	code << endl;
	code << "namespace ecg {" << endl;
	code << endl;

	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::messageTable" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const MSG_HOLDER " << GetName() << "_messageTable[] = {" << endl;
	for (MESSAGE_MAP::const_iterator iter = messages.begin(); iter != messages.end(); ++iter) {
		code << "	{ " << ConvertToLiteral(iter->second.ID()) << ", " << ConvertToLiteral(iter->second.Format()) << ", ";
		if (!iter->second.Info().empty()) {
			code << ConvertToLiteral(iter->second.Info());
		}
		code << " }," << endl;
	}
	code << "};" << endl;

	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* GetMessage()                                                          */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "const MSG_HOLDER * " << GetName() << "_GetMessage(const std::string & id)" << endl;
	code << "{" << endl;
	code << "	int index = ecg::FindMessage(id, " << GetName() << "_messageTable, " << (int)messages.size() << ");" << endl;
	code << "	if (index == -1) {" << endl;
	code << "		return NULL;" << endl;
	code << "	}" << endl;
	code << "	return &" << GetName() << "_messageTable[index];" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* GetMessageByIndex()                                                   */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/*************************************************************************/" << endl;
	code << "const MSG_HOLDER * " << GetName() << "_GetMessageByIndex(int index)" << endl;
	code << "{" << endl;
	code << "	return &" << GetName() << "_messageTable[index];" << endl;
	code << "}" << endl;
	code << endl;

	code << "} // namespace ecg" << endl;
}

//
// End-of-file ECG_GEN_COMPILER.CPP
//
