
/*************************************************************************/
/* ECG_INTSET.H                                               06.08.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_INTSET_H
#define ECG_INTSET_H

namespace ecg {

/*************************************************************************/
/* struct INTERVAL                                                       */
/*                                                                       */
/*************************************************************************/
template <class T>
struct INTERVAL {
	T				start;
	T				end;

	INTERVAL		(void) : start(), end() {}
	INTERVAL		(const T & start, const T & end) : start(start), end(end) {}

	bool			queryInter		(const INTERVAL<T> & a, INTERVAL<T> & res) const;
	int				querySub		(const INTERVAL<T> & a, INTERVAL<T> & res, INTERVAL<T> & res2) const;
	bool			operator <		(const INTERVAL<T> & a) const { return start < a.start || start == a.start && end < a.end; }
	bool			operator ==		(const INTERVAL<T> & a) const { return start == a.start && end == a.end; }
	bool			operator !=		(const INTERVAL<T> & a) const { return start != a.start || end != a.end; }
};

typedef INTERVAL<unsigned char> CHAR_INTERVAL;
typedef INTERVAL<int>			INT_INTERVAL;

template <class T>
std::ostream & operator << (std::ostream & output, const INTERVAL<T> & a);

/*************************************************************************/
/* struct INTSET                                                         */
/*                                                                       */
/*************************************************************************/
template <class T>
struct INTSET {
	typedef INTERVAL<T> INTERVAL_T;
	typedef INTSET<T>	INTSET_T;
	typedef std::vector<INTERVAL_T> INTERVAL_ARRAY;

	bool			negative;
	INTERVAL_ARRAY	intervals;

	INTSET			(void) : negative(false) {}
	INTSET			(const INTSET_T & a) : negative(a.negative), intervals(a.intervals) {}

	bool			empty			(void) const { return !negative && intervals.empty(); }
	void			clear			(void) { intervals.clear(); negative = false; }
	INTSET_T &		add				(const T & start, const T & end) { return add(INTERVAL_T(start, end)); }
	INTSET_T &		add				(const T & single) { return add(INTERVAL_T(single, single)); }
	INTSET_T &		add				(const INTERVAL_T & a) { intervals.push_back(a); return *this; }
	INTSET_T &		optimize		(void);

	INTSET_T &		operator +=		(const INTSET_T & a);
	INTSET_T &		operator *=		(const INTSET_T & a);
	INTSET_T &		operator -=		(const INTSET_T & a);
	INTSET_T &		negate			(void);

	bool			operator <		(const INTSET_T & a) const;
	bool			operator ==		(const INTSET_T & a) const;
	bool			operator !=		(const INTSET_T & a) const { return !(*this == a); }

	INTSET_T		operator +		(const INTSET_T & a) const;
	INTSET_T		operator *		(const INTSET_T & a) const;
	INTSET_T		operator -		(const INTSET_T & a) const;
	INTSET_T		operator ~		(void) const;

private:
	void			add				(const INTSET_T & a);
	void			mul				(const INTSET_T & a);
	void			sub				(const INTSET_T & a);
};

typedef INTSET<unsigned char>	CHAR_INTSET;
typedef INTSET<int>				INT_INTSET;

template <class T>
std::ostream & operator << (std::ostream & output, const INTSET<T> & a);

#include "ecg_intset.hpp"

};

#endif // ECG_INTSET_H
