
/*************************************************************************/
/* LEX_IMPL.H                                                        E13 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef LEX_IMPL_H
#define LEX_IMPL_H

#ifndef TOKEN_EOF
#define TOKEN_EOF		-1
#endif

#ifndef TOKEN_ERROR
#define TOKEN_ERROR		-2
#endif

namespace ecg {

class LEX;

/*************************************************************************/
/* class LEX_IMPL                                                        */
/*                                                                       */
/*************************************************************************/
class LEX_IMPL {
public:
	enum {
		STATE_COUNT		= 31,
		TERM_COUNT		= 14,
		NONTERM_COUNT	= 6,
		RULE_COUNT		= 19,
	};

	//enum TERMINALS {
	//		TOKEN_T_SLASH,			// "\"
	//		TOKEN_T_CHAR,			// $char
	//		TOKEN_T_NAME_1,			// "{"
	//		TOKEN_T_NAME_2,			// "}"
	//		TOKEN_T_NAME_3,			// "("
	//		TOKEN_T_NAME_4,			// ")"
	//		TOKEN_T_NAME_5,			// "["
	//		TOKEN_T_NAME_6,			// "]"
	//		TOKEN_T_NAME_7,			// "*"
	//		TOKEN_T_NAME_8,			// "+"
	//		TOKEN_T_NAME_9,			// "|"
	//		TOKEN_T_NAME_10,		// "^"
	//		TOKEN_T_NAME_11,		// "-"
	//		TOKEN_T_CHAR_SL,		// $char_sl
	//};

	enum NONTERMINALS {
		TOKEN_NT_PROGRAM,	// program
		TOKEN_NT_E,			// e
		TOKEN_NT_SET2,		// set2
		TOKEN_NT_SET,		// set
		TOKEN_NT_S,			// s
		TOKEN_NT_PROGRAM_1,	// program_1
	};

	typedef int VALUE_T;
	typedef LEX IMPL_T;

public:
						LEX_IMPL		(void);
	void				Compile				(void);

protected:
	void				intPushString		(void);
	void				intPushEmptyString	(void);
	void				intCreateStringArray(void);
	void				intAppendStringArray(void);

	template <class T> void intCreatePtrArray(void);
	template <class T> void intAppendPtrArray(void);
	template <class T> void intCreateValArray(void);
	template <class T> void intAppendValArray(void);

protected:
	void				AnyRegexp_Impl0		(void);
	void				CharsUnion_Impl0		(void);
	void				ConcatRegexp_Impl0		(void);
	void				FixedCharSet_Impl0		(void);
	void				FixedSetRegexp_Impl0		(void);
	void				ManyRegexp_Impl0		(void);
	void				NegativeSet_Impl0		(void);
	void				OrRegexp_Impl0		(void);
	void				RangeChars_Impl0		(void);
	void				RestorePreset_Impl0		(void);
	void				SetFinalRegexp_Impl0		(void);
	void				SetRegexp_Impl0		(void);
	void				SimpleRegexp_Impl0		(void);
	void				SingleChar_Impl0		(void);
	void				YesNotRegexp_Impl0		(void);
	void				intConstructor_Impl0		(void);

protected:
	COMPILER_STATE_F	compilerState;
	PTR_STACK			pstack;
	std::stack<std::string> sstack;
	RAW_STACK			vstack;

private:
	IMPL_T &			Impl			(void);

private:
	typedef AUTO_CONFIG<STATE_COUNT, TERM_COUNT, NONTERM_COUNT, RULE_COUNT> CONFIG_T;
	typedef AUTO<IMPL_T, CONFIG_T> AUTO_T;
	friend AUTO_T;

	typedef AUTO_T::ACTION		ACTION;
	typedef AUTO_T::ACTION_PROC	ACTION_PROC;
	typedef AUTO_T::PRODUCTION	PRODUCTION;

	static const ACTION			actionTable			[STATE_COUNT][TERM_COUNT];
	static const ACTION			actionTableEof		[STATE_COUNT];
	static const ACTION			actionTableError	[STATE_COUNT];
	static const int			productionTable		[STATE_COUNT][NONTERM_COUNT];
	static const PRODUCTION		procTableProduction	[RULE_COUNT];
	static const ACTION_PROC	procTableTerm		[TERM_COUNT];
	static const ACTION_PROC	procTableNonterm	[NONTERM_COUNT];

	AUTO_T				compilerAuto;
};

} // namespace ecg

#endif // LEX_IMPL_H

