
/*************************************************************************/
/* ECG_LEX_COMPILER.CPP                                       09.08.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "ecg_generator.h"

using namespace std;
using namespace ecg;

/*************************************************************************/
/* LEX::Generate()                                                       */
/*                                                                       */
/*************************************************************************/
void LEX::Generate(const std::string & name, const std::string & nameImpl, std::ostream & header, std::ostream & code, std::ostream & hlp, GENERATOR * generator)
{
	string nameUpr = name;
	strupr(&nameUpr[0]);

	// use skip if there are at least one preset with skipping
	bool useSkip = false;

	// use productive only if presets enabled and there are at least one nonproductive state
	bool useProductive = false;

	for (PRESET_ARRAY::const_iterator preset = presets.begin(); preset != presets.end(); ++preset) {
		if (preset->isSkip) {
			useSkip = true;
		}

		if (!useProductive && enablePresets) {
			for (int i=0; (size_t)i<preset->productive.size(); ++i) {
				if (preset->productive[i] == PRESET::PROD_NO) {
					useProductive = true;
				}
			}
		}
	}

	// use line counting only if char skipping enabled (it seem's not necessary to
	// count in solid text (without skipping) declaration (like lexer)
	bool useLineCounting = useSkip;

	header << endl;
	header << "/*************************************************************************/" << endl;
	header << "/* LEX                                                               E13 */" << endl;
	header << "/*                                                                       */" << endl;
	header << "/* generated by ecg                                                      */" << endl;
	header << "/*************************************************************************/" << endl;
	header << endl;
	header << "#ifndef " << nameUpr << "_H" << endl;
	header << "#define " << nameUpr << "_H" << endl;
	header << endl;
	header << "#ifndef TOKEN_EOF" << endl;
	header << "#define TOKEN_EOF		-1" << endl;
	header << "#endif" << endl;
	header << endl;
	header << "#ifndef TOKEN_ERROR" << endl;
	header << "#define TOKEN_ERROR		-2" << endl;
	header << "#endif" << endl;
	header << endl;
	header << "namespace ecg {" << endl;
	header << endl;
	header << "class " << name << " {" << endl;
	header << "public:" << endl;

	header << "	enum TERMINALS {" << endl;
	int maxlen = 0;
	for (GENTERM_ARRAY::const_iterator term = genTerms.begin(); term != genTerms.end(); ++term) {
		maxlen = max(maxlen, (int)term->alias.size() + 1);
	}
	for (GENTERM_ARRAY::const_iterator term = genTerms.begin(); term != genTerms.end(); ++term) {
		header << "		" << term->alias << "," << string(maxlen - (int)term->alias.size(), ' ') <<  "// '" << term->name << "'" << endl;
	}
	header << "	};" << endl;
	header << endl;

	bool enableLineComments		= commentLine != TOKEN_EOF;
	bool enableBlockComments	= commentBlockBegin != TOKEN_EOF && commentBlockEnd != TOKEN_EOF;
	bool enableComments			= enableLineComments || enableBlockComments;
	assert(enableBlockComments || commentBlockBegin == TOKEN_EOF && commentBlockEnd == TOKEN_EOF);

	if (enableComments) {
		header << "	enum {" << endl;
		if (enableLineComments) {
			assert(genTermsNmb.find(commentLine) != genTermsNmb.end());
			header << "		TOKEN_COMMENT_LINE			= " << genTerms[genTermsNmb[commentLine]].alias << "," << endl;
		}
		if (enableBlockComments) {
			assert(genTermsNmb.find(commentBlockBegin) != genTermsNmb.end());
			assert(genTermsNmb.find(commentBlockEnd)   != genTermsNmb.end());
			header << "		TOKEN_COMMENT_BLOCK_BEGIN	= " << genTerms[genTermsNmb[commentBlockBegin]].alias << "," << endl;
			header << "		TOKEN_COMMENT_BLOCK_END		= " << genTerms[genTermsNmb[commentBlockEnd]].alias   << "," << endl;
		}
		header << "	};" << endl;
		header << endl;
	}

	header << "	typedef unsigned char CHAR_T;" << endl;
	header << endl;
	header << "public:" << endl;
	header << "	" << name << "(void);" << endl;
	header << endl;
	header << "	int					Token			(void) const { return token; }" << endl;
	header << "	const std::string & TokenValue		(void) const { return value; }" << endl;

	if (useLineCounting) {
		header << "	int					Line			(void) const { return line + 1; }" << endl;
	}

	header << "	int 				NextToken		(void);" << endl;
	header << endl;
	
	if (enablePresets) {
		header << "	enum PRESET {" << endl;
		for (PRESET_ARRAY::iterator preset = presets.begin(); preset != presets.end(); ++preset) {
			preset->nameUpr = "PRESET_" + preset->name;
			assert(strlen(&preset->nameUpr[0]) == preset->nameUpr.size());
			strupr(&preset->nameUpr[0]);
			header << "		" << preset->nameUpr << "," << endl;
		}
		header << "	};" << endl;

		header << endl;
		header << "	void				SetPreset		(PRESET preset);" << endl;
		header << "	void				PushPreset		(PRESET preset);" << endl;
		header << "	void				PopPreset		(void);" << endl;
		header << "	PRESET				GetPreset		(void) const;" << endl;
		header << endl;
	}

	header << "protected:" << endl;
	header << "	virtual void		Unexpected		(char ch);" << endl;
	header << "	virtual void		Unexpected		(const std::string & val = std::string()); // empty string means end of file" << endl;
	header << endl;
	header << "private:" << endl;
	header << "	bool				PopChar		(CHAR_T & ch);" << endl;
	header << "	void				PushChar	(CHAR_T ch);" << endl;
	header << endl;
	header << "protected:" << endl;
	header << "	int					token;" << endl;
	header << "	std::string			value;" << endl;

	if (useLineCounting) {
		header << "	int					line;" << endl;
	}

	header << endl;
	header << "private:" << endl;
	header << "	// enables unget any count of chars" << endl;
	header << "	std::stack<CHAR_T>	charStack;" << endl;
	header << endl;

	if (enablePresets) {
		header << "	// preset data" << endl;
		header << "	std::stack<PRESET>	presetsStack;" << endl;
		header << endl;
		for (PRESET_ARRAY::const_iterator preset = presets.begin(); preset != presets.end(); ++preset) {
			if (useSkip) {
				header << "	static const int 	isSkip_" << preset->name << "[];" << endl;
			}
			header << "	static const int 	final_" << preset->name << "[];" << endl;

			if (useProductive) {
				header << "	static const int 	productive_" << preset->name << "[];" << endl;
			}
		}
		header << endl;
		header << "	// parsing data" << endl;
		header << "	const int *			final;" << endl;

		if (useSkip) {
			header << "	const int *			isSkip;" << endl;
		}
		if (useProductive) {
			header << "	const int *			productive;" << endl;
		}
	} else {
		header << "	// parsing data" << endl;
		header << "	static const int 	final[];" << endl;
		if (useSkip) {
			header << "	static const int 	isSkip[];" << endl;
		}
	}
	header << endl;

	header << "	static const int	table[][" << (int)terminals.size() << "];" << endl;
	header << "	static const int	terminal[];" << endl;
	header << "};" << endl;
	header << endl;
	header << "} // namespace ecg" << endl;
	header << endl;
	header << "#endif // " << nameUpr << "_H" << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< CODE >>>

	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* LEX                                                               E13 */" << endl;
	code << "/*                                                                       */" << endl;
	code << "/* generated by ecg                                                      */" << endl;
	code << "/*************************************************************************/" << endl;
	code << endl;
	code << "#include \"stdafx.h\"" << endl;
	code << "#include \"" << generator->implName << ".h\"" << endl;
	code << endl;
	code << "using namespace std;" << endl;
	code << "using namespace ecg;" << endl;
	code << endl;

	// TABLE
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::table" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const int " << name << "::table[" << (int)table.size() << "][" << (int)terminals.size() << "] = {" << endl;
	for (STATE_ARRAY::const_iterator line = table.begin(); line != table.end(); ++line) {
		code << "	{ ";
		for (int i=0; (size_t)i<terminals.size(); ++i) {
			INT_INT_MAP::const_iterator edge = (*line)->edges.find(i);
			if (edge != (*line)->edges.end()) {
				code << edge->second << ", ";
			} else {
				code << "-1, ";
			}
		}
		code << "}," << endl;
	}
	code << "};" << endl;
	code << endl;

	// TERMINALS
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::terminal" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "const int " << name << "::terminal[" << (UCHAR_MAX + 1) << "] = { ";
	CHARS ch;
	ch.add(0);
	for (int i = 0; i < UCHAR_MAX + 1; ++i) {
		if (i % 32 == 0) {
			code << endl <<	"	";
		}

		ch.intervals[0].start = (char)i;
		ch.intervals[0].end	  = (char)i;
		bool found = false;
		for (int t = 0; (size_t)t < terminals.size(); ++t) {
			if (!(terminals[t] * ch).empty()) {
				code << t << ", ";
				found = true;
				break;
			}
		}
		if (!found) {
			code << "-1, ";
		}
	}
	code << endl << "};" << endl;
	code << endl;

	//////////////////////////////////////////////////////////////////////////
	// <<< PRESETS >>>
	for (PRESET_ARRAY::const_iterator preset = presets.begin(); preset != presets.end(); ++preset) {
		string presetName = preset->name;

		// FINAL STATES
		string finalName = !enablePresets ? "final" : "final_" + preset->name;
		code << "/*************************************************************************/" << endl;
		code << "/* " << name << "::" << finalName << endl;
		code << "/* " << endl;
		code << "/*************************************************************************/" << endl;
		code << "const int " << name << "::" << finalName << "[" << (int)table.size() << "] = {" << endl;
		code << "	";
		for (INT_ARRAY::const_iterator token = preset->final.begin(); token != preset->final.end(); ++token) {
			if (*token == TOKEN_EOF) {
				code << "TOKEN_EOF, ";
			} else {
				code << genTerms[genTermsNmb[*token]].alias << ", ";
			}
		}
		code << endl << "};" << endl;
		code << endl;

		if (useSkip) {
			// SKIP SYMBOLS
			string skipName = !enablePresets ? "isSkip" : "isSkip_" + preset->name;
			code << "/*************************************************************************/" << endl;
			code << "/* " << name << "::" << skipName << endl;
			code << "/* " << endl;
			code << "/*************************************************************************/" << endl;
			code << "const int " << name << "::" << skipName << "[" << (UCHAR_MAX + 1) << "] = {";
			for (int i = 0; i < UCHAR_MAX + 1; ++i) {
				if (i % 32 == 0) {
					code << endl <<	"	";
				}

				ch.intervals[0].start = (char)i;
				ch.intervals[0].end	  = (char)i;
				if ((preset->skip * ch).empty()) {
					code << "0, ";
				} else {
					code << "1, ";
				}
			}
			code << endl << "};" << endl;
			code << endl;
		}

		if (useProductive) {
			// PRODUCTIVE SYMBOLS
			string productiveName = "productive_" + preset->name;
			code << "/*************************************************************************/" << endl;
			code << "/* " << name << "::" << productiveName << endl;
			code << "/* " << endl;
			code << "/*************************************************************************/" << endl;
			code << "const int " << name << "::" << productiveName << "[" << (int)table.size() << "] = {" << endl;
			code << "	";
			for (PRESET::PRODUCTIVE_ARRAY::const_iterator prod = preset->productive.begin(); prod != preset->productive.end(); ++prod) {
				assert(*prod == PRESET::PROD_YES || *prod == PRESET::PROD_NO);
				if (*prod == PRESET::PROD_YES) {
					code << "1, ";
				} else {
					code << "0, ";
				}
			}
			code << endl << "};" << endl;
			code << endl;
		}
	}

	// PRESET SELECT
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::" << name << "()" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << name << "::" << name << "(void)" << endl;
	// dummy string to use ':' and constructor in any generation condition
	code << "	: charStack()" << endl;

	if (useLineCounting) {
		code << "	, line()" << endl;
	}

	if (enablePresets) {
		code << "	";
		if (useSkip) {
			code << ", isSkip(isSkip_default)";
		}
		code << ", final(final_default)" << endl;
		code << "{" << endl;
		code << "	presetsStack.push(PRESET_DEFAULT);" << endl;
		code << "	SetPreset(PRESET_DEFAULT);" << endl;
		code << "}" << endl;
	} else {
		code << "{}" << endl;
	}
	code << endl;

	if (enablePresets) {
		code << "/*************************************************************************/" << endl;
		code << "/* " << name << "::SetPreset()" << endl;
		code << "/* " << endl;
		code << "/*************************************************************************/" << endl;
		code << "void " << name << "::SetPreset(PRESET preset)" << endl;
		code << "{" << endl;
		code << "	assert(!presetsStack.empty());" << endl;
		code << "	presetsStack.top() = preset;" << endl;
		code << endl;
		code << "	switch (preset) {" << endl;
		
		for (PRESET_ARRAY::const_iterator preset = presets.begin(); preset != presets.end(); ++preset) {
			code << "	case " << preset->nameUpr << ":" << endl;
			code << "		final		= final_" << preset->name << ";" << endl;
			if (useSkip) {
				code << "		isSkip		= isSkip_" << preset->name << ";" << endl;
			}
			if (useProductive) {
				code << "		productive	= productive_" << preset->name << ";" << endl;
			}
			code << "		break;" << endl;
		}

		code << "	default:" << endl;
		code << "		assert(false);" << endl;
		code << "		break;" << endl;
		code << "	}" << endl;
		code << "}" << endl;
		code << endl;
		code << "/*************************************************************************/" << endl;
		code << "/* " << name << "::PushPreset()" << endl;
		code << "/* " << endl;
		code << "/*************************************************************************/" << endl;
		code << "void " << name << "::PushPreset(PRESET preset)" << endl;
		code << "{" << endl;
		code << "	assert(!presetsStack.empty());" << endl;
		code << "	presetsStack.push(preset);" << endl;
		code << "	SetPreset(preset);" << endl;
		code << "}" << endl;
		code << endl;
		code << "/*************************************************************************/" << endl;
		code << "/* " << name << "::PopPreset()" << endl;
		code << "/* " << endl;
		code << "/*************************************************************************/" << endl;
		code << "void " << name << "::PopPreset(void)" << endl;
		code << "{" << endl;
		code << "	assert(!presetsStack.empty());" << endl;
		code << "	presetsStack.pop();" << endl;
		code << "	SetPreset(presetsStack.top());" << endl;
		code << "}" << endl;
		code << endl;
		code << "/*************************************************************************/" << endl;
		code << "/* " << name << "::GetPreset()" << endl;
		code << "/* " << endl;
		code << "/*************************************************************************/" << endl;
		code << name << "::PRESET " << name << "::GetPreset(void) const" << endl;
		code << "{" << endl;
		code << "	assert(!presetsStack.empty());" << endl;
		code << "	return presetsStack.top();" << endl;
		code << "}" << endl;
		code << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// <<< LEX PARSE >>>

	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::PopChar()" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "bool " << name << "::PopChar(CHAR_T & ch)" << endl;
	code << "{" << endl;
	code << "	// check queue" << endl;
	code << "	if (!charStack.empty()) {" << endl;
	code << "		ch = charStack.top();" << endl;
	code << "		charStack.pop();" << endl;
	code << "	} else if (!static_cast<" << nameImpl << " *> (this)->GetChar(ch)) {" << endl;
	code << "		// end of file" << endl;
	code << "		return false;" << endl;
	code << "	}" << endl;
	code << endl;

	if (useLineCounting) {
		code << "	// line counter" << endl;
		code << "	if (ch == '\\n') {" << endl;
		code << "		++line;" << endl;
		code << "	}" << endl;
		code << endl;
	}

	code << "	return true;" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::PushChar()" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "void " << name << "::PushChar(CHAR_T ch)" << endl;
	code << "{" << endl;

	if (useLineCounting) {
		code << "	// check for eoln" << endl;
		code << "	if (ch == '\\n') {" << endl;
		code << "		--line;" << endl;
		code << "	}" << endl;
		code << endl;
	}

	code << "	// push char to queue" << endl;
	code << "	charStack.push(ch);" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::Unexpected()" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "void " << name << "::Unexpected(const std::string & val)" << endl;
	code << "{" << endl;
	code << "	token = TOKEN_ERROR;" << endl;
	code << "	value = val;" << endl;
	code << "}" << endl;
	code << endl;
	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::Unexpected()" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "void " << name << "::Unexpected(char ch)" << endl;
	code << "{" << endl;
	code << "	token = TOKEN_ERROR;" << endl;
	code << "	value = ch;" << endl;
	code << "}" << endl;
	code << endl;

	code << "/*************************************************************************/" << endl;
	code << "/* " << name << "::NextToken()" << endl;
	code << "/* " << endl;
	code << "/*************************************************************************/" << endl;
	code << "int " << name << "::NextToken(void)" << endl;
	code << "{" << endl;
	code << "	// current char" << endl;
	code << "	CHAR_T ch;" << endl;
	code << endl;

	if (enableComments) {
		code << "repeat_comment:" << endl;
		code << endl;
	}

	if (useSkip) {
		code << "	// skip chars" << endl;
		code << "	for (;;) {" << endl;
		code << "		// get next char" << endl;
		code << "		if (!PopChar(ch)) {" << endl;
		code << "			break;" << endl;
		code << "		}" << endl;
		code << endl;
		code << "		// check skipable" << endl;
		code << "		if (!isSkip[ch]) {" << endl;
		code << "			PushChar(ch);" << endl;
		code << "			break;" << endl;
		code << "		}" << endl;
		code << "	}" << endl;
		code << endl;
	}

	code << "	// auto state" << endl;
	code << "	int state = 0;" << endl;
	code << endl;
	code << "	// clear value" << endl;
	code << "	value.clear();" << endl;
	code << endl;
	code << "	// result" << endl;
	code << "	int good = -1;" << endl;
	code << endl;
	code << "	for (;;) {" << endl;
	code << "		// check for final state (first state can be final too)" << endl;
	code << "		if (final[state] != TOKEN_EOF) {" << endl;
	code << "			// check if terminal needs break symbol" << endl;
	code << "			good = (int)value.size();" << endl;
	code << "			token = final[state];" << endl;
	code << "		}" << endl;
	code << endl;
	code << "		// get next char" << endl;
	code << "		if (!PopChar(ch)) {" << endl;
	code << "			break;" << endl;
	code << "		}" << endl;
	code << endl;
	code << "		// get terminal" << endl;
	code << "		int term = terminal[ch];" << endl;
	code << endl;
	code << "		// check unknown char" << endl;
	code << "		if (term < 0) {" << endl;    
	code << "			Unexpected(ch);" << endl;
	code << "			return token;" << endl;		
	code << "		}" << endl;					
	code << endl;
	code << "		// value" << endl;
	code << "		value += ch;" << endl;
	code << endl;
	code << "		// next state" << endl;                  
	code << "		int next = table[state][term];" << endl;
	code << endl;
	code << "		// check end of parse" << endl;
	code << "		if (next < 0";

	if (useProductive) {
		code << " || !productive[next]";
	}

	code << ") {" << endl;
	code << "			break;" << endl;
	code << "		}" << endl;
	code << endl;
	code << "		// move to next state" << endl;
	code << "		state = next;" << endl;
	code << "	}" << endl;
	code << endl;
	code << "	// check for end of file" << endl;
	code << "	if (good == -1) {" << endl;
	code << "		// check for unexpected" << endl;
	code << "		if (!value.empty()) {" << endl;
	code << "			Unexpected(value);" << endl;
	code << "			return token;" << endl;
	code << "		}" << endl;
	code << endl;
	code << "		token = TOKEN_EOF;" << endl;
	code << "		value = \"<EOF>\";" << endl;
	code << "		return token;" << endl;
	code << "	}" << endl;
	code << endl;
	code << "	// unget some chars" << endl;
	code << "	for (int i=(int)value.size()-1; i >= good; --i) {" << endl;
	code << "		PushChar(value[i]);" << endl;
	code << "	}" << endl;
	code << "	value.resize(good);" << endl;
	code << endl;

	if (enableLineComments) {
		code << "	// check line comments" << endl;
		code << "	if (token == TOKEN_COMMENT_LINE) {" << endl;
		code << "		// skip all chars to end of line" << endl;
		code << "		for (;;) {" << endl;
		code << "			if (!PopChar(ch)) {" << endl;
		code << "				break;" << endl;
		code << "			}" << endl;
		code << "			if (ch == '\\n') {" << endl;
		code << "				break;" << endl;
		code << "			}" << endl;
		code << "		}" << endl;
		code << "		goto repeat_comment;" << endl;
		code << "	}" << endl;
		code << endl;
	}

	if (enableBlockComments) {
		code << "	// check block comment" << endl;
		code << "	if (token == TOKEN_COMMENT_BLOCK_BEGIN) {" << endl;
		code << "		// unget all chars, it allows blocks like /*/" << endl;
		code << "		for (int i=(int)value.size()-1; i >= 0; --i) {" << endl;
		code << "			PushChar(value[i]);" << endl;
		code << "		}" << endl;
		code << endl;
		code << "		// set special preset" << endl;
		code << "		PushPreset(PRESET_COMMENT);" << endl;
		code << endl;
		code << "		// initial state" << endl;
		code << "		state = 0;" << endl;
		code << endl;
		code << "		for (;;) {" << endl;
		code << "			// check unexpected end of file" << endl;
		code << "			if (!PopChar(ch)) {" << endl;
		code << "				PopPreset();" << endl;
		code << "				Unexpected("");" << endl;
		code << "				return token;" << endl;
		code << "			}" << endl;
		code << endl;
		code << "			// get next state" << endl;
		code << "			int term  = terminal[ch];" << endl;
		code << "			assert(term >= 0);" << endl;
		code << "			state = table[state][term];" << endl;
		code << "			assert(state >= 0);" << endl;
		code << endl;
		code << "			// check end of comment block" << endl;
		code << "			if (final[state] == TOKEN_COMMENT_BLOCK_END) {" << endl;
		code << "				PopPreset();" << endl;
		code << "				goto repeat_comment;" << endl;
		code << "			}" << endl;
		code << endl;
		code << "			assert(final[state] == TOKEN_EOF);" << endl;
		code << "		}" << endl;
		code << "	}" << endl;
		code << endl;
	}

	code << "	// OK!!! Value and token contain valid info" << endl;
	code << "	return token;" << endl;
	code << "}" << endl;
	
	// HELP FOR PROGRAMMERS
	set<int>   markNode;
	queue<int> nextNode;
	hlp << endl << "<<< GRAPHS >>>" << endl << endl;
	for (INT_SET::const_iterator node = startNodes.begin(); node != startNodes.end(); ++node) {
		nextNode.push(*node);
		markNode.insert(*node);
		hlp << "--------------------------------------------" << endl;
		while (!nextNode.empty()) {
			int node = nextNode.front();
			nextNode.pop();

			hlp << "NODE #" << node << endl;
			for (EDGE_ARRAY::const_iterator edge = nodes[node].edges.begin(); edge != nodes[node].edges.end(); ++edge) {
				hlp << "	" << edge->next << ": " << edge->cond << endl;
				if (markNode.find(edge->next) == markNode.end()) {
					nextNode.push(edge->next);
					markNode.insert(edge->next);
				}
			}
			if (nodes[node].token != TOKEN_EOF) {
				hlp << "	final : " << genTerms[genTermsNmb[nodes[node].token]].name << endl;
				hlp << "	prior : " << nodes[node].prior << endl;
			}
		}
	}
	hlp << "--------------------------------------------" << endl;

	hlp << endl << "<<< TABLE >>>" << endl << endl;

	for (int line = 0; (size_t)line < table.size(); ++line) {
		hlp << "STATE #" << line << endl;
		hlp << "	{ ";
		for (INT_SET::const_iterator iter = table[line]->nodes.begin(); iter != table[line]->nodes.end(); ++iter) {
			hlp << *iter;
			++iter;
			if (iter != table[line]->nodes.end()) {
				hlp << ", ";
			}
			--iter;
		}
		hlp << " }" << endl;

		for (INT_INT_MAP::const_iterator iter = table[line]->edges.begin(); iter != table[line]->edges.end(); ++iter) {
			hlp << "	" << iter->second << " : " << terminals[iter->first] << endl;
		}

		if (table[line]->token != TOKEN_EOF) {
			hlp << "	final : " << genTerms[genTermsNmb[table[line]->token]].name << endl;
		}
	}
}

//
// End-of-file ECG_LEX_COMPILER.CPP
//
