
/*************************************************************************/
/* ECG_LOADER_IMPL.CPP                                      29.07.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "ecg_generator.h"

using namespace std;
using namespace ecg;

/*************************************************************************/
/* LOADER::GetChar()                                                     */
/*                                                                       */
/*************************************************************************/
bool LOADER::GetChar(CHAR_T & ch)
{
	char_traits<char>::int_type ich = input->get();
	ch = char_traits<char>::to_char_type(ich);
	return ich >= 0;
}

/*************************************************************************/
/* LOADER::Init()                                                        */
/*                                                                       */
/*************************************************************************/
void LOADER::Init(GENERATOR & generator, istream & input, ostream & error)
{
	this->input = &input;
	this->error = &error;
	this->generator = &generator;
}

/*************************************************************************/
/* LOADER::TypePtr()                                                     */
/*                                                                       */
/*************************************************************************/
void LOADER::Type(VALUE_TYPE valueType, PARAM_TYPE paramType, const std::string & type, P_TYPE & res)
{
	res	= new TYPE(type, paramType == WORD, valueType == VAL, false);

	if (paramType == WORD && (type == ecg::VOID_TYPE || type == ecg::STRING_TYPE)) {
		if (valueType == PTR) {
			messageTable["name"] = type;
			FatalError(MESSAGE_LDR1031);
			return ;
		}

		assert(valueType == VAL);
		res->isImplMember = false;
	}
}

/*************************************************************************/
/* LOADER::TypeArray()                                                   */
/*                                                                       */
/*************************************************************************/
void LOADER::TypeArray(VALUE_TYPE valueType, PARAM_TYPE paramType, const std::string & type, P_TYPE & res)
{
	if (paramType == WORD && type == ecg::VOID_TYPE) {
		FatalError(MESSAGE_LDR1015);
		return ;
	}

	res = new TYPE(type, paramType == WORD, valueType == VAL, true);
}

/*************************************************************************/
/* LOADER::TYPE::ToStrType()                                             */
/*                                                                       */
/*************************************************************************/
string LOADER::TYPE::ToString(void) const
{
	string res = name;

	if (!isValue) {
		res = "[" + res + "]";
	}
	if (isArray) {
		res += "*";
	}
	return res;
}

/*************************************************************************/
/* LOADER::TYPE::ToStrType()                                             */
/*                                                                       */
/*************************************************************************/
string LOADER::TYPE::ToStrType(void) const
{
	if (isArray) {
		if (name == ecg::STRING_TYPE) {
			return "std::vector<std::string>";
		}

		if (isValue) {
			return "std::vector<" + (isImplMember ? "IMPL_T::" + name : name) + " >";
		} else {
			return "std::vector<P<" + (isImplMember ? "IMPL_T::" + name : name) + " > >";
		}
	}
	
	return (isImplMember ? "IMPL_T::" + name : name);
}

/*************************************************************************/
/* LOADER::LiteralToString()                                        
/*                                                                       
/* worl			-> "L" $literal;									<<< LiteralToString >>>
/*************************************************************************/
void LOADER::LiteralToString(const std::string & lit, std::string & res)
{
	// process 'word' to '"word"'
	res = "\"" + lit + "\"";
}

/*************************************************************************/
/* LOADER::FindNontermType()                                             */
/*                                                                       */
/*************************************************************************/
LOADER::P_TYPE LOADER::FindNontermType(P_ITEM nonterm, ITEM_SET & typeSearching, bool & cycleFound)
{
	if (typeSearching.find(nonterm) != typeSearching.end()) {
		// return empty string, it means that found cycle
		cycleFound = true;
		return new TYPE(ecg::VOID_TYPE);
	}

	// assume that nonterm has no type yet
	assert(typedItems.find(nonterm) == typedItems.end());

	// mark searching this nonterm's type
	typeSearching.insert(nonterm);

	assert(nonterm->type == NONTERMINAL);
	ecg::NONTERMINAL & nt = generator->nonterminal[nonterm->id];

	// by default curType is empty (it means "not defined")
	P_TYPE curType = new TYPE();

	RULE_INDEX index;
	index.nontermId = nt.id;
	for (int ruleId = 0; (size_t)ruleId < nt.rules.size(); ++ruleId) {
		// check common type for this rule
		P_TYPE ruleType = new TYPE();
		P_TYPE type = new TYPE();

		index.ruleNumber = ruleId;
		assert(ruleCode.find(index) != ruleCode.end());
		CODE_DATA::ARRAY_SUGG arraySugg = ruleCode[index].arraySugg;

		ecg::RULE::ITEM_ARRAY & items = nt.rules[ruleId].items;
		for (ecg::RULE::ITEM_ARRAY::const_iterator item = items.begin(); item != items.end(); ++item) {
			if (item->isTerm) {
				P_ITEM term(new ITEM(TERMINAL, item->id));
				assert(typedItems.find(term) != typedItems.end());
				type = typedItems[term];
			} else {
				P_ITEM nontermItem(new ITEM(NONTERMINAL, item->id));

				if (typedItems.find(nontermItem) == typedItems.end()) {
					typedItems[nontermItem] = FindNontermType(nontermItem, typeSearching, cycleFound);
				}

				type = typedItems[nontermItem];
			}

			// check for cycle
			if (cycleFound && arraySugg == CODE_DATA::NONE) {
				typeSearching.erase(nonterm);
				return new TYPE(ecg::VOID_TYPE);
			}

			// if arraySugg not NONE - ignore cycle
			cycleFound = false;

			// skip void types
			if (*type == ecg::VOID_TYPE) {
				continue;
			}

			if (ruleType->IsEmpty()) {
				// if first non-void type, set it to result type
				ruleType = type;
			} else {
				// rule has more then one args, so its result type is void
				typeSearching.erase(nonterm);
				return new TYPE(ecg::VOID_TYPE);
			}
		}

		// if ruleType is empty = all args has void type, so just skip it
		if (ruleType->IsEmpty()) {
			continue;
		}

		if (curType->IsEmpty()) {
			// set current type
			// for now doesn't have array of array
			if (arraySugg == CODE_DATA::NONE || ruleType->isArray) {
				curType = ruleType;
			} else {
				curType = new TYPE(ruleType->name, ruleType->isImplMember, ruleType->isValue, true);
			}
		} else {
			// compare types with checking for array
			P_TYPE cmpType = ruleType;
			if (arraySugg != CODE_DATA::NONE && !ruleType->isArray) {
				cmpType = new TYPE(ruleType->name, ruleType->isImplMember, ruleType->isValue, true);
			}

			if (*curType != *cmpType) {
				// if different rule type - result type is void
				typeSearching.erase(nonterm);
				return new TYPE(ecg::VOID_TYPE);
			}
		}
	}

	// curType is empty - it means only that this nonterm has productions
	// only with empty arg-lists, so this type is void
	if (curType->IsEmpty()) {
		curType = new TYPE(ecg::VOID_TYPE);
	}

	// unmark searching this nonterm's type
	typeSearching.erase(nonterm);

	// wow! how long path for this line!
	return curType;
}

/*************************************************************************/
/* LOADER::PrintMessage()                                                */
/*                                                                       */
/*************************************************************************/
void LOADER::PrintMessage(int msgId, int line /* = -1 */)
{
	string msg = LOADER_GetMessageByIndex(msgId)->format;
	msg = "Line {line}: " + msg;
	messageTable["line"] = ToString((line == -1) ? Line() : line);
	GetError() << Compose(msg, messageTable) << endl;
}

/*************************************************************************/
/* LOADER::FatalError()                                                  */
/*                                                                       */
/*************************************************************************/
void LOADER::FatalError(int messageIndex, int line)
{
	PrintMessage(messageIndex, line);
	compilerState.set(CST_FATAL_ERROR);
}

/**************************************************************************
/* LOADER::SetName
/*
/* program			-> "#name" "(" worl "," worl ")" declaration_list; <<< CommitGrammar >>>
/*************************************************************************/
void LOADER::CommitGrammar(const std::string & nameImpl)
{
	// compiler name
	generator->nameImpl	= nameImpl;

	// compiler's implementation name
   assert(!nameImpl.empty());
   if (islower(char_traits<char>::to_int_type(nameImpl[nameImpl.size() - 1]))) {
      generator->name = nameImpl + "Impl";
   } else {
      generator->name = nameImpl + "_IMPL";
   }

	//////////////////////////////////////////////////////////////////////////
	// <<< TYPES >>>

	// first, check terminals
	// fill all nontyped terminals by void
	P_ITEM tmItem;
	for (ecg::TERMINAL_ARRAY::const_iterator tm = generator->terminal.begin(); tm != generator->terminal.end(); ++tm) {
		tmItem = new ITEM(TERMINAL, tm->id);
		if (typedItems.find(tmItem) == typedItems.end()) {
			typedItems[tmItem] = new TYPE(ecg::VOID_TYPE);
		}
	}

	// $error has type void
	tmItem = new ITEM(TERMINAL, TOKEN_ERROR);
	assert(typedItems.find(tmItem) == typedItems.end());
	typedItems[tmItem] = new TYPE(ecg::VOID_TYPE);

	ITEM_SET typeSearching;

	// lets check nonterminals
	for (ecg::NONTERMINAL_ARRAY::const_iterator nt = generator->nonterminal.begin(); nt != generator->nonterminal.end(); ++nt) {
		P_ITEM ntItem (new ITEM(NONTERMINAL, nt->id));
		if (typedItems.find(ntItem) == typedItems.end()) {
			// if autotype disabled - set nonterminal's type to void
			if (!enableAutotype) {
				typedItems[ntItem] = new TYPE(ecg::VOID_TYPE);
			} else {
				bool cycleFound = false;
				typedItems[ntItem] = FindNontermType(ntItem, typeSearching, cycleFound);
			}
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// <<< CODE BLOCKS >>>

	CODE_BLOCK cb;
	string code;
	for (TYPED_ITEM_MAP::const_iterator ti = typedItems.begin(); ti != typedItems.end(); ++ti) {
		P_ITEM item = ti->first;
		P_TYPE type = ti->second;

		if (item->type == TERMINAL) {
			// skip $error terminal
			if (item->id == TOKEN_ERROR) {
				continue;
			}

			// register terminal's code
			if (!RetrieveFunctionName(item, code)) {
				compilerState.set(CST_FATAL_ERROR);
				return ;
			}
			assert((size_t)item->id < generator->terminal.size());
			assert(generator->terminal[item->id].code.empty());
			generator->terminal[item->id].code = code;
		} else {
			assert(item->type == NONTERMINAL);

			// register nonterminal's code
			if (!RetrieveFunctionName(item, code)) {
				compilerState.set(CST_FATAL_ERROR);
				return ;
			}
			assert((size_t)item->id < generator->nonterminal.size());
			assert(generator->nonterminal[item->id].code.empty());
			generator->nonterminal[item->id].code = code;

			// process rules
			ecg::RULE_ARRAY & rules = generator->nonterminal[item->id].rules;

			RULE_INDEX index;
			index.nontermId = item->id;

			// check each rule
			for (int ruleId = 0; (size_t)ruleId < rules.size(); ++ruleId) {
				index.ruleNumber = ruleId;

				if (!RetrieveFunctionName(index, code)) {
					compilerState.set(CST_FATAL_ERROR);
					return ;
				}

				assert(rules[ruleId].code.empty());
				rules[ruleId].code = code;
			}
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// <<< LEX >>>

	if (generator->enableLex) {
		for (TERMINAL_ARRAY::const_iterator iter = generator->terminal.begin(); iter != generator->terminal.end(); ++iter) {
			assert(lex.genTermsNmb.find(iter->id) != lex.genTermsNmb.end());

			// check terminals without regexp
			if (!lex.genTerms[lex.genTermsNmb[iter->id]].hasRegexp) {
				assert(!iter->name.empty());
				if (needRegexp.find(iter->id) != needRegexp.end()) {
					messageTable["name"] = iter->name;
					FatalError(MESSAGE_LDR1001);
					return ;
				}

				// if terminal not need's regexp -> register it with own name like "name" -> regexp("name")
				string regexp;
				for (string::const_iterator ch = iter->name.begin(); ch != iter->name.end(); ++ch) {
					if (!isalpha(char_traits<char>::to_int_type(*ch))) {
						regexp += '\\';
					} 
					regexp += *ch;
				}

				lex.Init(iter->id, regexp, LEX::PRIORITY_NONE, Line(), GetError());
				lex.Compile();
				assert(lex.compilerState.value() == 0);
			}
		}

		if (!lex.CalcTable()) {
			FatalError(MESSAGE_LDR1002);
			return ;
		}
	}
}

/**************************************************************************
/* LOADER::SetAlias
/*
/* declaration		-> param "=" term;									<<< SetAlias >>>
/*************************************************************************/
void LOADER::SetAlias(P_PARAM alias, P_PARAM param)
{
	assert(!alias->IsTerminal());

	// check for simple alias
	if (!param->IsTerminal()) {
		messageTable["name"] = param->value;
		FatalError(MESSAGE_LDR1003);
		return ;
	}

	// terminal-alias
	generator->aliasTA.insert(make_pair(param->value, alias->value));

	// add terminal (must be added after aliasTA contains alias)
	int termId = AddTerminal(param);

	// alias-terminal
	generator->aliasAT.insert(make_pair(alias->value, termId));
}

/*************************************************************************/
/* LOADER::GetItemName()                                                 */
/*                                                                       */
/*************************************************************************/
string LOADER::GetItemName(P_ITEM item) const
{
	if (item->type == NONTERMINAL) {
		assert((size_t)item->id < generator->nonterminal.size());
		return generator->nonterminal[item->id].name;
	}

	assert(item->type == TERMINAL);

	if (item->id == TOKEN_ERROR) {
		return "$error";
	}

	assert((size_t)item->id < generator->terminal.size());
	return generator->terminal[item->id].name;
}

/**************************************************************************
/* LOADER::DeclTerminals
/*
/* declaration		-> (param % "|") code_block;	<<< SetItemsCodeBlock >>>
/*************************************************************************/
void LOADER::SetItemsCodeBlock(P<PARAM_LIST> itemList, P<CODE_BLOCK> cb)
{
	assert(!cb->name.empty());

	for (PARAM_LIST::const_iterator item = itemList->begin(); item != itemList->end(); ++item) {
		// check for terminal
		P_ITEM tm;

		// value can be terminal's alias (not nonterminal)
		STRING_INT_MAP::const_iterator alias = generator->aliasAT.find((*item)->value);

		if ((*item)->IsTerminal()) {
			int termId = AddTerminal(*item);
			tm.reset(new ITEM(TERMINAL, termId));
		} else if (alias != generator->aliasAT.end()) {
			tm.reset(new ITEM(TERMINAL, alias->second));
		} else {
			generator->AddNonterminal((*item)->value);
			tm.reset(new ITEM(NONTERMINAL, generator->nonterminalNumber[(*item)->value]));
		}

		// check non-empty code block
		if (itemCode.find(tm) != itemCode.end()) {
			GetError() << "WARNING: Replacing code block for item \"" << (*item)->value << "\"" << endl;
			GetError() << "> Old: <<< " << itemCode[tm].code->ToString() << " >>>" << endl;
			GetError() << "> New: <<< " << cb->ToString() << " >>>" << endl;
			GetError() << endl;
		}

		CODE_DATA cd;
		cd.code = cb;
		cd.line = Line();
		itemCode[tm] = cd;
	}
}

/*************************************************************************/
/* LOADER::SimpleCodeBlock()                                      
/*                                                                       
/* code_block_ne	-> "<<<" worl ">>>";			<<< SimpleCodeBlock >>>
/*************************************************************************/
void LOADER::SimpleCodeBlock(const std::string & code, P<CODE_BLOCK> & cb)
{
	cb = new CODE_BLOCK(code, code.empty());
}

/**************************************************************************
/* LOADER::CreateRule
/*
/* rule			-> $word "->" rule_or_list nonterm_type;			<<< CreateRule >>>
/*************************************************************************/
void LOADER::CreateRule(const std::string & nonterm, P<RULE_OR_LIST> ruleOrList, P_TYPE nontermType, P<RULE> & rule)
{
	// check for incorrect name
	assert(!nonterm.empty());
	if (nonterm[0] == '$') {
		messageTable["name"] = nonterm;
		FatalError(MESSAGE_LDR1004);
		return ;
	}

	// register nonterminal
	generator->AddNonterminal(nonterm);
	int nontermId = generator->nonterminalNumber[nonterm];

	rule.reset(new RULE());
	rule->nontermId = nontermId;

	// add rules and fill rule->productions
	for (RULE_OR_LIST::const_iterator ruleItem = ruleOrList->begin(); ruleItem < ruleOrList->end(); ++ruleItem) {
		ProcessNewRule(nontermId, (*ruleItem)->type, (*ruleItem)->itemList, (*ruleItem)->itemList2, rule->productions);
	}

	// set nonterm result type
	if (!nontermType->IsEmpty()) {
		typedItems.insert(make_pair(new ITEM(NONTERMINAL, nontermId), nontermType));
	}
}

/*************************************************************************/
/* LOADER::ProcessNewRule()                                            */
/*                                                                       */
/*************************************************************************/
int LOADER::ProcessNewRule(int nontermId, RULE_ITEM_TYPE type, P<ITEM_LIST> itemList, P<ITEM_LIST> itemList2, INT_SET & productions)
{
	ecg::NONTERMINAL & nt = generator->nonterminal[nontermId];

	P<ITEM_LIST> tmpList = new ITEM_LIST();

	// returns number greater then -1 only in SIMPLE rule because multiple rules
	int ruleAdded = -1;

	switch (type) {
		case SIMPLE:
			{
				assert(itemList2 == NULL);

				ecg::RULE	rule;
				if (itemList != NULL) {
					for (ITEM_LIST::const_iterator item = itemList->begin(); item < itemList->end(); ++item) {
						ecg::RULE::ITEM tm;
						tm.isTerm	= (*item)->type == TERMINAL;
						tm.id		= (*item)->id;
						rule.items.push_back(tm);
					}
				}

				int ruleId = generator->FindRule(nt.rules, rule);
				if (ruleId == -1) {
					ruleId = (int)nt.rules.size();
					nt.rules.push_back(rule);
				}

				// check markers to determine rule number where it used
				if (itemList != NULL) {
					for (ITEM_LIST::const_iterator item = itemList->begin(); item < itemList->end(); ++item) {
						if ((*item)->type == NONTERMINAL && marker.find((*item)->id) != marker.end()) {
							assert(marker[(*item)->id].nontermId == -1 && marker[(*item)->id].ruleId == -1);
							marker[(*item)->id].nontermId	= nontermId;
							marker[(*item)->id].ruleId		= ruleId;
						}
					}
				}

				// update ruleAdded with added rule :)
				ruleAdded = ruleId;

				// store production number
				productions.insert(ruleId);

				// register rule line number
				RULE_INDEX index;
				index.nontermId		= nontermId;
				index.ruleNumber	= ruleId;
				if (ruleCode.find(index) == ruleCode.end()) {
					CODE_DATA cdata;
					cdata.line = Line();
					ruleCode[index] = cdata;
				}
			}
			break;

		case PERIODIC:
			{
				assert(itemList != NULL && itemList2 == NULL);

				RULE_INDEX index;
				index.nontermId = nontermId;

				// A -> ;
				index.ruleNumber = ProcessNewRule(nontermId, SIMPLE, NULL, NULL, productions);
				assert(index.ruleNumber >= 0);
				assert(ruleCode.find(index) != ruleCode.end());

				// don't set any array suggestion because constructors

				// A -> A b;
				tmpList->clear();
				tmpList->push_back(new ITEM(NONTERMINAL, nontermId));
				copy(itemList->begin(), itemList->end(), back_inserter(*tmpList));

				index.ruleNumber = ProcessNewRule(nontermId, SIMPLE, tmpList, NULL, productions);
				assert(index.ruleNumber >= 0);
				assert(ruleCode.find(index) != ruleCode.end());
				ruleCode[index].arraySugg = CODE_DATA::APPEND_ARRAY;
			}
			break;

		case NEM_PERIODIC:
			{
				assert(itemList != NULL && itemList2 == NULL);

				RULE_INDEX index;
				index.nontermId = nontermId;

				// A -> b;
				index.ruleNumber = ProcessNewRule(nontermId, SIMPLE, itemList, NULL, productions);
				assert(index.ruleNumber >= 0);
				assert(ruleCode.find(index) != ruleCode.end());
				ruleCode[index].arraySugg = CODE_DATA::INIT_ARRAY;

				// A -> A b;
				tmpList->clear();
				tmpList->push_back(new ITEM(NONTERMINAL, nontermId));
				copy(itemList->begin(), itemList->end(), back_inserter(*tmpList));
				
				index.ruleNumber = ProcessNewRule(nontermId, SIMPLE, tmpList, NULL, productions);
				assert(index.ruleNumber >= 0);
				assert(ruleCode.find(index) != ruleCode.end());
				ruleCode[index].arraySugg = CODE_DATA::APPEND_ARRAY;
			}
			break;

		case DEL_PERIODIC:
			{
				assert(itemList != NULL && itemList2 != NULL);

				RULE_INDEX index;
				index.nontermId = nontermId;

				// A -> b;
				index.ruleNumber = ProcessNewRule(nontermId, SIMPLE, itemList, NULL, productions);
				assert(index.ruleNumber >= 0);
				assert(ruleCode.find(index) != ruleCode.end());
				ruleCode[index].arraySugg = CODE_DATA::INIT_ARRAY;

				// A -> A c b;
				tmpList->clear();
				// A
				tmpList->push_back(new ITEM(NONTERMINAL, nontermId));
				// c
				copy(itemList2->begin(), itemList2->end(), back_inserter(*tmpList));
				// b
				copy(itemList->begin(), itemList->end(), back_inserter(*tmpList));
				
				index.ruleNumber = ProcessNewRule(nontermId, SIMPLE, tmpList, NULL, productions);
				assert(index.ruleNumber >= 0);
				assert(ruleCode.find(index) != ruleCode.end());
				ruleCode[index].arraySugg = CODE_DATA::APPEND_ARRAY;
			}
			break;

		case YESNOT:
			{
				assert(itemList != NULL && itemList2 == NULL);

				// A -> ;
				ProcessNewRule(nontermId, SIMPLE, NULL, NULL, productions);

				// A -> b;
				ProcessNewRule(nontermId, SIMPLE, itemList, NULL, productions);
			}
			break;

		default:
			assert(false);
			break;
	}

	return ruleAdded;
}

/**************************************************************************
/* LOADER::AddTerminalItem
/*
/* item			-> $literal;					<<< AddTerminalItem >>>
/* term			-> $literal;					<<< AddTerminalItem >>>
/*************************************************************************/
void LOADER::CreateTermItem(const std::string & term, P_ITEM & termItem)
{
	// register new
	int termId = AddTerminal(new PARAM(LITERAL, term));

	termItem = new ITEM(TERMINAL, termId);
}

/**************************************************************************
/* LOADER::CreateWordItem
/*
/* item			-> $word;						<<< CreateWordItem(false) >>>
/* term			-> $word;						<<< CreateWordItem(true) >>>
/*************************************************************************/
void LOADER::CreateWordItem(bool onlyTerminal, const std::string & name, P_ITEM & item)
{
	// name: terminal or alias or nonterminal

	// try to find in aliases
	STRING_INT_MAP::const_iterator iter = generator->aliasAT.find(name);
	if (iter != generator->aliasAT.end()) {
		item = new ITEM(TERMINAL, iter->second);
		return ;
	}

	// name can't be empty
	assert(!name.empty());

	// if name begins from '$' this is terminal
	if (name[0] == '$') {
		int termId = AddTerminal(new PARAM(WORD, name));
		item.reset(new ITEM(TERMINAL, termId));
	} else {
		if (onlyTerminal) {
			messageTable["name"] = name;
			FatalError(MESSAGE_LDR1005);
			return ;
		}

		generator->AddNonterminal(name);
		item = new ITEM(NONTERMINAL, generator->nonterminalNumber[name]);
	}
}

/**************************************************************************
/* LOADER::CreateBracketsItem
/*
/* item			-> "(" rule ")";				<<< CreateBracketsItem >>>
/*************************************************************************/
void LOADER::CreateBracketsItem(P<RULE> rule, P_ITEM & item)
{
	item = new ITEM(NONTERMINAL, rule->nontermId);
}

/**************************************************************************
/* LOADER::CreateBracketsOrItem
/*
/* item			-> "(" rule_or_list nonterm_type ")";		<<< CreateBracketsOrItem >>>
/*************************************************************************/
void LOADER::CreateBracketsOrItem(P<RULE_OR_LIST> ruleOrList, P<TYPE> nontermType, P_ITEM & item)
{
	// create anonymous nonterminal
	string nonterm = generator->MakeAnonymousNontermName();
	generator->AddNonterminal(nonterm);
	int nontermId = generator->nonterminalNumber[nonterm];

	// productions
	INT_SET productions;

	// add rules and fill rule->productions
	for (RULE_OR_LIST::const_iterator ruleItem = ruleOrList->begin(); ruleItem < ruleOrList->end(); ++ruleItem) {
		ProcessNewRule(nontermId, (*ruleItem)->type, (*ruleItem)->itemList, (*ruleItem)->itemList2, productions);
	}

	P_ITEM rule = new ITEM(NONTERMINAL, nontermId);

	// nonterm type
	if (!nontermType->name.empty()) {
		typedItems.insert(make_pair(rule, nontermType));
	}

	// push rule result
	item = rule;
}

/**************************************************************************
/* LOADER::CreateCodeBlockItem
/*
/* item			-> code_block;								<<< CreateCodeBlockItem >>>
/*************************************************************************/
void LOADER::CreateCodeBlockItem(P<CODE_BLOCK> cb, P_ITEM & item)
{
	// create marker nonterminal
	string nonterm = generator->MakeAnonymousNontermName();
	generator->AddNonterminal(nonterm);
	int nontermId = generator->nonterminalNumber[nonterm];

	// M -> ;
	INT_SET productions;
	ProcessNewRule(nontermId, SIMPLE, NULL, NULL, productions);

	assert(productions.size() == 1);
	assert(*productions.begin() == 0);
	assert(generator->nonterminal[nontermId].rules.size() == 1);

	// set code for this marker
	assert(!cb->name.empty());
	
	RULE_INDEX index;
	index.nontermId		= nontermId;
	index.ruleNumber	= 0;
	ruleCode[index].code = cb;
	ruleCode[index].line = Line();

	// register marker for params detection
	marker.insert(make_pair(nontermId, MARKER()));

	// push result
	item = new ITEM(NONTERMINAL, nontermId);
}

/*************************************************************************/
/* LOADER::CreateYesNotItem()                                          
/*                                                                       
/* item			-> "[" rule_item "]"; 								<<< CreateYesNotItem >>>
/*************************************************************************/
void LOADER::CreateYesNotItem(P<RULE_ITEM> ruleItem, P_ITEM & item)
{
	// make anonymous nonterminal
	string nonterm = generator->MakeAnonymousNontermName();
	generator->AddNonterminal(nonterm);
	int nontermId = generator->nonterminalNumber[nonterm];

	// productions
	INT_SET productions;

	// A -> ;
	ProcessNewRule(nontermId, SIMPLE, NULL, NULL, productions);

	// A -> a;
	ProcessNewRule(nontermId, ruleItem->type, ruleItem->itemList, ruleItem->itemList2, productions);

	item = new ITEM(NONTERMINAL, nontermId);
}

/*************************************************************************/
/* LOADER::ParametrizedCodeBlock()                                
/*                                                                       
/* code_block_ne	-> "<<<" worl params_ne ">>>";  <<< ParametrizedCodeBlock >>>
/*************************************************************************/
void LOADER::ParametrizedCodeBlock(const std::string & code, P<PARAM_LIST> params, P<CODE_BLOCK> & cb)
{
	STRING_ARRAY args;
	for (PARAM_LIST::const_iterator iter = params->begin(); iter != params->end(); ++iter) {
		string value = (*iter)->value;
		if ((*iter)->fromWord) {
			if (value != "true" && value != "false" && !isdigit(char_traits<char>::to_int_type(value[0]))) {
				value = "IMPL_T::" + value;
			}
		}
		args.push_back(value);
	}

	cb = new CODE_BLOCK(code, code.empty(), args);
}

/*************************************************************************/
/* operator ==()                                                         */
/*                                                                       */
/*************************************************************************/
bool LOADER::CODE_BLOCK::operator == (const CODE_BLOCK & cb) const
{
	return isConstructor == cb.isConstructor
		&& name == cb.name
		&& ecg::AUTO_FUNCTION::CompareArrays(params, cb.params) == 0;
}

/*************************************************************************/
/* LOADER::CODE_BLOCK::ToString()                                        */
/*                                                                       */
/*************************************************************************/
string LOADER::CODE_BLOCK::ToString(void) const
{
	ostringstream res;
	res << name;
	if (!params.empty()) {
		res << "(";
		for (STRING_ARRAY::const_iterator param = params.begin(); param < params.end(); ++param) {
			res << *param;
			if (param + 1 < params.end()) {
				res << ", ";
			}
		}
		res << ")";
	}

	return res.str();
}

/*************************************************************************/
/* LOADER::PrintTypedRule()                                              */
/*                                                                       */
/*************************************************************************/
void LOADER::PrintTypedRule(std::ostream & output, P_TYPE result, const P_TYPE_ARRAY & params, bool printEoln) const
{
	output << "[";
	if (result->IsEmpty()) {
		output << "void";
	} else {
		output << result->ToString();
	}
	output << " -> ";
	for (P_TYPE_ARRAY::const_iterator param = params.begin(); param < params.end(); ++param) {
		output << (*param)->ToString();
		if (param + 1 < params.end()) {
			output << " ";
		}
	}
	output << "]";
	
	if (printEoln) {
		output << endl;
	}
}

/*************************************************************************/
/* LOADER::RetrieveFunctionName()                                        */
/*                                                                       */
/*************************************************************************/
bool LOADER::RetrieveFunctionName(P_ITEM item, std::string & code)
{
	P_TYPE type;
	if (item->type == TERMINAL) {
		assert(typedItems.find(item) != typedItems.end());
		type = typedItems[item];
	} else {
		assert(item->type == NONTERMINAL);
		type.reset(new TYPE(ecg::VOID_TYPE));
	}

	// check when code block can be empty
	assert(!type->IsEmpty());
	bool canBeEmpty = *type == ecg::VOID_TYPE || *type == ecg::STRING_TYPE;

	CODE_DATA codeData;
	if (itemCode.find(item) != itemCode.end()) {
		codeData = itemCode[item];
	} else {
		codeData.code = new CODE_BLOCK();
		codeData.line = Line();
	}

	if (codeData.code->name.empty()) {
		if (canBeEmpty) {
			if (*type == ecg::STRING_TYPE) {
				code = ecg::PUSH_STRING_METHOD;
			} else {
				code.clear();
			}
			return true;
		}

		// error: empty code block
		PrintMessage(MESSAGE_LDR1006, codeData.line);
		GetError() << "> Terminal = " << GetItemName(item) << endl;
		GetError() << "> Type     = " << type->ToString() << endl;
		GetError() << endl;

		return false;
	}

	// check for processing function void ()
	if (codeData.code->params.empty() && *type == ecg::VOID_TYPE) {
		code = codeData.code->name;
		return true;
	}

	// create autofunction for terminals
	ecg::AUTO_FUNCTION af;
	af.name			= codeData.code->name;
	af.constParams	= codeData.code->params;

	af.result.isValue	= type->isValue && !type->isArray;
	af.result.name		= type->ToStrType();
	af.params.clear();

	code = generator->RegisterAutoFunction(af);

	return true;
}

/*************************************************************************/
/* LOADER::GetTypesUntil()                                               */
/*                                                                       */
/*************************************************************************/
void LOADER::GetTypesUntil(P_TYPE_ARRAY & params, int nontermId, int ruleId, int untilNontermId /*= -1*/)
{
	ecg::RULE::ITEM_ARRAY & items = generator->nonterminal[nontermId].rules[ruleId].items;

	for (ecg::RULE::ITEM_ARRAY::const_iterator item = items.begin(); item != items.end(); ++item) {
		// if marker found - stop type search
		if (!item->isTerm && item->id == untilNontermId) {
			break;
		}

		// find item's type
		P_ITEM tm = new ITEM(item->isTerm ? TERMINAL : NONTERMINAL, item->id);
		assert(typedItems.find(tm) != typedItems.end());
		P_TYPE type = typedItems[tm];

		// skip items without types
		assert(!type->IsEmpty());
		if (*type != ecg::VOID_TYPE) {
			params.push_back(type);
		}
	}
}

/*************************************************************************/
/* LOADER::RetrieveFunctionName()						                      */
/*                                                                       */
/*************************************************************************/
bool LOADER::RetrieveFunctionName(const RULE_INDEX & ruleIndex, string & code)
{
	// retrieve rule result type
	assert(typedItems.find(new ITEM(NONTERMINAL, ruleIndex.nontermId)) != typedItems.end());
	P_TYPE result = typedItems[new ITEM(NONTERMINAL, ruleIndex.nontermId)];

	// retrieve rule params
	P_TYPE_ARRAY params;
	GetTypesUntil(params, ruleIndex.nontermId, ruleIndex.ruleNumber);

	bool isMarkerFunc = false;

	// check for marker
	if (marker.find(ruleIndex.nontermId) != marker.end()) {
		assert(params.empty());

		// get marker's params
		MARKER & m = marker[ruleIndex.nontermId];
 		GetTypesUntil(params, m.nontermId, m.ruleId, ruleIndex.nontermId);

		// mark marker func
		isMarkerFunc = true;
	}

	// check when code block can be empty
	bool canBeEmpty =
		params.empty() && *result == ecg::VOID_TYPE ||
		params.size() == 1 && *result == *params[0];

	assert(ruleCode.find(ruleIndex) != ruleCode.end());
	CODE_DATA codeData = ruleCode[ruleIndex];

	// code can be empty
	if (codeData.code == NULL) {
		codeData.code = new CODE_BLOCK();
	}

	bool isConstructorFunc = codeData.code->isConstructor;

	// check code block empty
	if (codeData.code->name.empty() && !isConstructorFunc) {
		if (canBeEmpty) {
			code.clear();
			return true;
		}

		// if result type string and empty params - just push empty string in the stack
		if (*result == ecg::STRING_TYPE && params.empty()) {
			code = ecg::PUSH_EMPTY_STRING_METHOD;
			return true;
		}

		// check for InitArray
		if (codeData.arraySugg == CODE_DATA::INIT_ARRAY
		&&  result->isArray
		&&  params.size() == 1
		&&  !params[0]->isArray
		&&  result->name == params[0]->name) {
			if (*params[0] == ecg::STRING_TYPE) {
				assert(params[0]->isValue);
				code = ecg::INIT_STRING_ARRAY_METHOD;
			} else if (params[0]->isValue) {
				code = ecg::INIT_VAL_ARRAY_METHOD_T + ("<" + params[0]->ToStrType() + " >");
			} else {
				code = ecg::INIT_PTR_ARRAY_METHOD_T + ("<" + params[0]->ToStrType() + " >");
			}
			return true;
		}

		// check for AppendArray
		if (codeData.arraySugg == CODE_DATA::APPEND_ARRAY
		&&  result->isArray
		&&  params.size() == 2
		&&  *result == *params[0]
		&&  !params[1]->isArray
		&&  result->name == params[1]->name) {
			if (*params[1] == ecg::STRING_TYPE) {
				assert(params[0]->isValue);
				code = ecg::APPEND_STRING_ARRAY_METHOD;
			} else if (params[1]->isValue) {
				code = ecg::APPEND_VAL_ARRAY_METHOD_T + ("<" + params[1]->ToStrType() + " >");
			} else {
				code = ecg::APPEND_PTR_ARRAY_METHOD_T + ("<" + params[1]->ToStrType() + " >");
			}
			return true;
		}

		// can't use constructor with 'void', but make sure that there is no markers
		if (*result == ecg::VOID_TYPE) {
			ecg::RULE::ITEM_ARRAY & items = generator->nonterminal[ruleIndex.nontermId].rules[ruleIndex.ruleNumber].items;

			// check if all args were used in markers
			bool error = false;
			for (ecg::RULE::ITEM_ARRAY::reverse_iterator iter = items.rbegin(); iter != items.rend(); ++iter) {
				// if marker found - all args used in this one
				if (!iter->isTerm && marker.find(iter->id) != marker.end()) {
					break;
				}

				// check item type
				P_ITEM tm = new ITEM(iter->isTerm ? TERMINAL : NONTERMINAL, iter->id);
				if (*typedItems[tm] != ecg::VOID_TYPE) {
					error = true;
					break;
				}
			}

			if (error) {
				// error: empty code block
				PrintMessage(MESSAGE_LDR1007, codeData.line);
				GetError() << "[";
				generator->PrintRule(GetError(), generator->nonterminal[ruleIndex.nontermId].name, items);
				GetError() << "]" << endl;

				PrintTypedRule(GetError(), result, params);
				GetError() << endl;

				return false;
			}
		}

		isConstructorFunc = true;
	}

	// check when autofunction not needed
	// autofunc not needed if we want to construct object that already exists
	if (!isConstructorFunc && canBeEmpty && params.empty() && codeData.code->params.empty()
	||	isConstructorFunc && params.size() == 1 && *result == *params[0]) {
		code		= codeData.code->name;
		return true;
	}

	// constructor couldn't be applied to value type (only 'void' and 'string')
	if (isConstructorFunc && result->isValue && *result != ecg::STRING_TYPE && *result != ecg::VOID_TYPE) {
		PrintMessage(MESSAGE_LDR1032, codeData.line);

		GetError() << "[";
		ecg::RULE::ITEM_ARRAY & items = generator->nonterminal[ruleIndex.nontermId].rules[ruleIndex.ruleNumber].items;
		generator->PrintRule(GetError(), generator->nonterminal[ruleIndex.nontermId].name, items);
		GetError() << "]" << endl;

		PrintTypedRule(GetError(), result, params);
		GetError() << endl;

		return false;
	}

	// register autofunction
	ecg::AUTO_FUNCTION af;
	assert(isConstructorFunc == codeData.code->name.empty());
	af.isConstructor	= isConstructorFunc;
	af.isMarker			= isMarkerFunc;
	af.name				= codeData.code->name;
	af.constParams		= codeData.code->params;
	af.result.isValue	= result->isValue;
	af.result.name		= result->ToStrType();
	for (P_TYPE_ARRAY::const_iterator param = params.begin(); param != params.end(); ++param) {
		AUTO_FUNCTION::PARAM funcParam;
		funcParam.name		= (*param)->ToStrType();
		funcParam.isValue	= (*param)->isValue && !(*param)->isArray;
		af.params.push_back(funcParam);
	}

	code = generator->RegisterAutoFunction(af);

	return true;
}

/*************************************************************************/
/* LOADER::CommitRuleAndItsCodeBlock()                            
/*
/* declaration		-> rule ";" attrib_list; <<< CommitRuleAndItsCodeBlock >>>
/*************************************************************************/
void LOADER::CommitRuleAndItsCodeBlock(P<RULE> rule, P<RULE_ATTRIB> attrib)
{
	ecg::NONTERMINAL & nt = generator->nonterminal[rule->nontermId];

	// check priority
	if (!attrib->prior.empty() && priorityAlias.find(attrib->prior) == priorityAlias.end()) {
		messageTable["name"] = attrib->prior;
		FatalError(MESSAGE_LDR1008);
		return ;
	}

	PRIORITY prior;
	if (!attrib->prior.empty()) {
		prior = priorityAlias[attrib->prior];
	} else {
		assert(prior.priority == -1);
	}

	for (INT_SET::const_iterator prod = rule->productions.begin(); prod != rule->productions.end(); ++prod) {
		RULE_INDEX index;
		index.nontermId		= nt.id;
		index.ruleNumber	= *prod;

		// <<< CODE BLOCK >>>
		if (attrib->codeBlock != NULL) {
			// register code block
			RULE_CODE_MAP::const_iterator cdIter = ruleCode.find(index);
			if (cdIter != ruleCode.end()) {
				if (cdIter->second.code != NULL && *cdIter->second.code != *attrib->codeBlock) {
					messageTable["name"] = nt.name;
					PrintMessage(MESSAGE_LDR1009);
					GetError() << "[";
					generator->PrintRule(GetError(), nt.name, nt.rules[*prod].items);
					GetError() << "]" << endl;
					GetError() << "> Old: <<< " << nt.rules[*prod].code << " >>>" << endl;
					GetError() << "> New: <<< " << attrib->codeBlock->ToString() << " >>>" << endl;
					GetError() << endl;
				}
			}

			ruleCode[index].code = attrib->codeBlock;
			ruleCode[index].line = Line();
		}

		// <<< PRIORITY >>>
		if (prior.priority != -1) {
			if (!generator->RegisterRulePriority(index, prior.left, prior.priority)) {
				PrintMessage(MESSAGE_LDR1010);
				GetError() << "[";
				generator->PrintRule(GetError(), index.nontermId, index.ruleNumber);
				GetError() << "]" << endl;
				GetError() << "> Priority: " << attrib->prior << endl;

				compilerState.set(CST_FATAL_ERROR);
				return ;
			}
		}
	}

	// <<< PREFER >>>
	if (attrib->preferShift) {
		if (!RegisterPrefer(rule, attrib->specShift, true)) {
			compilerState.set(CST_FATAL_ERROR);
			return ;
		}
	}
	if (attrib->preferReduce) {
		if (!RegisterPrefer(rule, attrib->specReduce, false)) {
			compilerState.set(CST_FATAL_ERROR);
			return ;
		}
	}
}

/*************************************************************************/
/* LOADER::RegisterPrefer()                                              */
/*                                                                       */
/*************************************************************************/
bool LOADER::RegisterPrefer(P<RULE> rule, P<ITEM_LIST> itemList, bool preferShift)
{
	// make terminals list
	INT_SET spec;
	for (ITEM_LIST::const_iterator itSpec = itemList->begin(); itSpec < itemList->end(); ++itSpec) {
		spec.insert((*itSpec)->id);
	}

	// tmpPref for warnings
	INT_SET tmpSpec;
	for (INT_SET::const_iterator prod = rule->productions.begin(); prod != rule->productions.end(); ++prod) {
		if (generator->GetPrefer(rule->nontermId, *prod, tmpSpec, preferShift)) {
			PrintMessage(preferShift ? MESSAGE_LDR1011 : MESSAGE_LDR1012);
			GetError() << "> ";
			generator->PrintRule(GetError(), generator->nonterminal[rule->nontermId].name, generator->nonterminal[rule->nontermId].rules[*prod].items, true);
			GetError() << endl;

			compilerState.set(CST_FATAL_ERROR);
			return false;
		}

		generator->RegisterPrefer(rule->nontermId, *prod, spec, preferShift);
	}

	return true;
}

/*************************************************************************/
/* LOADER::RegisterResultType()                                     
/*                                                                       
/* declaration		-> item_list2 ":" type ";";						<<< RegisterResultType >>>
/*************************************************************************/
void LOADER::RegisterResultType(P<ITEM_LIST> itemList, P_TYPE type)
{
	for (ITEM_LIST::const_iterator pitem = itemList->begin(); pitem < itemList->end(); ++pitem) {
		P_ITEM item = *pitem;

		if (typedItems.find(item) == typedItems.end()) {
			typedItems.insert(make_pair(item, type));
			continue;
		}

		if (*typedItems[item] == *type) {
			continue;
		}

		messageTable["name"] = GetItemName(item);
		if (item->type == TERMINAL) {
			PrintMessage(MESSAGE_LDR1013);
		} else {
			assert(item->type == NONTERMINAL);
			PrintMessage(MESSAGE_LDR1014);
		}
		assert(!typedItems[item]->IsEmpty());
		assert(!type->IsEmpty());
		GetError() << "> Old: " << typedItems[item]->ToString() << endl;
		GetError() << "> New: " << type->ToString() << endl;
		GetError() << endl;

		compilerState.set(CST_FATAL_ERROR);
	}

	return ;
}

/*************************************************************************/
/* LOADER::ProcessLiteral()                                              */
/*                                                                       */
/*************************************************************************/
void LOADER::ProcessLiteral(std::string & res)
{
	string a = TokenValue();

	res.clear();
	for (int i = 0; (size_t)i < a.size(); ++i) {
		if (a[i] == '\\') {
			assert((size_t)i + 1 < a.size());
			++i;
		}

		res += a[i];
	}

	assert(res.size() >= 2);
	res = res.substr(1, res.size() - 2);
}

/*************************************************************************/
/* LOADER::ShiftAttrib()                                            
/*                                                                       
/* attrib_list		-> attrib_list "shift" term_params;					<<< ShiftAttrib >>>
/*************************************************************************/
void LOADER::ShiftAttrib(P_RULE_ATTRIB attrib, P<ITEM_LIST> params, P_RULE_ATTRIB & res)
{
	if (attrib->preferShift) {
		FatalError(MESSAGE_LDR1016);
		return ;
	}

	res = attrib;
	res->preferShift	= true;
	res->specShift		= params;
}

/*************************************************************************/
/* LOADER::ReduceAttrib()                                                */
/*                                                                       */
/*************************************************************************/
void LOADER::ReduceAttrib(P_RULE_ATTRIB attrib, P<ITEM_LIST> params, P_RULE_ATTRIB & res)
{
	if (attrib->preferReduce) {
		FatalError(MESSAGE_LDR1017);
		return ;
	}

	res = attrib;
	res->preferReduce	= true;
	res->specReduce		= params;
}

/*************************************************************************/
/* LOADER::CodeAttrib()                                             
/*                                                                       
/* attrib_list		-> attrib_list code_block; <<< CodeAttrib >>>
/*************************************************************************/
void LOADER::CodeAttrib(P_RULE_ATTRIB attrib, P_CODE_BLOCK params, P_RULE_ATTRIB & res)
{
	if (attrib->codeBlock != NULL) {
		messageTable["name"] = attrib->codeBlock->ToString();
		FatalError(MESSAGE_LDR1018);
		return ;
	}

	res = attrib;
	res->codeBlock = params;
}

/*************************************************************************/
/* LOADER::PriorAttrib()                                                 */
/*                                                                       */
/*************************************************************************/
void LOADER::PriorAttrib(P_RULE_ATTRIB attrib, const std::string prior, P_RULE_ATTRIB & res)
{
	if (!attrib->prior.empty()) {
		messageTable["name"] = attrib->prior;
		FatalError(MESSAGE_LDR1019);
		return ;
	}

	res = attrib;
	res->prior = prior;
}

/*************************************************************************/
/* LOADER::RegisterPriority()
/* 
/* declaration		-> "#left" (param +) $eoln;		<<< RegisterPriority(true) >>>
/* declaration		-> "#right"  (param +) $eoln;	<<< RegisterPriority(false) >>>
/*************************************************************************/
void LOADER::RegisterPriority(bool left, P<PARAM_LIST> terms)
{
	for (PARAM_LIST::const_iterator iter = terms->begin(); iter != terms->end(); ++iter) {
		if ((*iter)->IsTerminal()) {
			AddTerminal(*iter);

			if (!generator->RegisterOpPriority(generator->terminalNumber[(*iter)->value], left, maxPriority)) {
				messageTable["name"] = (*iter)->value;
				FatalError(MESSAGE_LDR1020);
				return ;
			}
		} else {
			if (priorityAlias.find((*iter)->value) != priorityAlias.end()) {
				messageTable["name"] = (*iter)->value;
				FatalError(MESSAGE_LDR1021);
				return ;
			}
			priorityAlias[(*iter)->value] = PRIORITY(left, maxPriority);
		}
	}

	++maxPriority;
}

/*************************************************************************/
/* PARAM::PARAM()                                                        */
/*                                                                       */
/*************************************************************************/
LOADER::PARAM::PARAM(LOADER::PARAM_TYPE type, const string & value)
	: value(value)
{
	if (type == WORD) {
		fromWord = true;
		return ;
	}

	fromWord = false;

	if (type == STRING) {
		this->value = "\"" + value + "\"";
		return ;
	}

	assert(type == LITERAL);
}

/*************************************************************************/
/* LOADER::LexRegexp()                                              
/*                                                                       
/* declaration		-> alias_name 	  "=" $regexp [":" {SPP} $r_prior]; <<< LexRegexp >>>
/*************************************************************************/
void LOADER::LexRegexp(P_PARAM name, const string & regexp, const string & prior)
{
	generator->enableLex = true;

	// must be terminal because 'alias_name'
	assert(name->IsTerminal());

	// register terminal (don't mark terminal needs regexp, because it has)
	int termId = AddTerminal(name);

	// select user defined priority
	LEX::PRIORITY userPriority = LEX::PRIORITY_NONE;
	if (prior == "low") {
		userPriority = LEX::PRIORITY_LOW;
	} else if (prior == "normal") {
		userPriority = LEX::PRIORITY_NORMAL;
	} else if (prior == "high") {
		userPriority = LEX::PRIORITY_HIGH;
	} else if (!prior.empty()) {
		messageTable["name"] = prior;
		FatalError(MESSAGE_LDR1022);
		return ;
	}

	lex.Init(termId, regexp, userPriority, Line(), GetError());
	lex.Compile();
	if (lex.compilerState.value() != 0) {
		messageTable["name"] = regexp;
		FatalError(MESSAGE_LDR1023);
		return ;
	}
}

/*************************************************************************/
/* LOADER::LexSkip()                                                     */
/*                                                                       */
/*************************************************************************/
void LOADER::LexSkip(const string & skip)
{
	generator->enableLex = true;

	lex.Init(Line(), GetError());

	LEX::CHARS s;
	if (!lex.StringToChars(skip.c_str(), (int)skip.size(), s)) {
		compilerState.set(CST_FATAL_ERROR);
		return ;
	}

	if (!lex.defSkip.empty()) {
		FatalError(MESSAGE_LDR1024);
		return ;
	}

	lex.defSkip = s;
}

/*************************************************************************/
/* LOADER::LexCommand()                                             
/*                                                                       
/* pr_com			-> "+" param;										<<< LexCommand(true) >>>
/* pr_com			-> "-" param;										<<< LexCommand(false) >>>
/*************************************************************************/
void LOADER::LexCommand(bool enable, P_PARAM param)
{
	int token = TOKEN_EOF;
	// register terminal if needed
	if (param->IsTerminal()) {
		token = AddTerminal(param);
	}

	// add new command
	assert((size_t)curPresetId < lex.presets.size());
	lex.presets[curPresetId].AddCommand(enable, param->value, token);
}

/*************************************************************************/
/* LOADER::LexPresetSkip()                                               */
/*                                                                       */
/*************************************************************************/
void LOADER::LexPresetSkip(const string & skip)
{
	lex.Init(Line(), GetError());

	// convert from string to character set
	LEX::CHARS s;
	if (!lex.StringToChars(skip.c_str(), (int)skip.size(), s)) {
		compilerState.set(CST_FATAL_ERROR);
		return ;
	}

	// check redefinition
	if (lex.presets[curPresetId].isSkip) {
		FatalError(MESSAGE_LDR1024);
		return ;
	}

	lex.presets[curPresetId].skip = s;
	lex.presets[curPresetId].isSkip = true;
}

/*************************************************************************/
/* LOADER::PopPresetString()                                             */
/*                                                                       */
/*************************************************************************/
void LOADER::PopPresetString(string & res)
{
	res = TokenValue();
	PopPreset();
}

/*************************************************************************/
/* LOADER::CheckAliasName()                                              */
/*                                                                       */
/*************************************************************************/
void LOADER::CheckAliasName(P_PARAM name, P_PARAM & res)
{
	res = name;

	// for nonterminal, just result its value - its alias ($word = param)
	if (!name->IsTerminal()) {
		return ;
	}

	// switch to regexp ($word = \D\S*)
	PushPreset(PRESET_REGEXP);
}

/*************************************************************************/
/* LOADER::LexCLine()                                               
/*                                                                       
/* declaration		-> "cline"	$regexp $eoln;			<<< LexCLine >>>
/*************************************************************************/
void LOADER::LexCLine(const std::string & cline)
{
	generator->enableLex = true;

	// register line comment terminal
	int termId = AddTerminal(new PARAM(WORD, COMMENT_LINE));
	lex.RegisterComment(COMMENT_PRESET, termId, COMMENT_LINE, generator->terminal[termId].alias);

	// enable line comment generation
	lex.commentLine = termId;

	// line comment redefinition - is not error
	lex.Init(termId, cline, LEX::PRIORITY_NONE, Line(), GetError(), false);
	lex.Compile();
	if (lex.compilerState.value() != 0) {
		FatalError(MESSAGE_LDR1025);
		return ;
	}
}

/*************************************************************************/
/* LOADER::LexCBegin()                                              
/*                                                                       
/* declaration		-> "cbegin" $regexp;								<<< LexCBegin >>>
/*************************************************************************/
void LOADER::LexCBegin(const std::string & cbegin)
{
	generator->enableLex = true;

	// register line comment terminal
	int termId = AddTerminal(new PARAM(WORD, COMMENT_BLOCK_BEGIN));
	lex.RegisterComment(COMMENT_PRESET, termId, COMMENT_BLOCK_BEGIN, generator->terminal[termId].alias);

	// enable block comment generation
	lex.commentBlockBegin = termId;

	// line comment redefinition - is not error
	lex.Init(termId, cbegin, LEX::PRIORITY_NONE, Line(), GetError(), false);
	lex.Compile();
	if (lex.compilerState.value() != 0) {
		FatalError(MESSAGE_LDR1025);
		return ;
	}
}

/*************************************************************************/
/* LOADER::LexCEnd()                                                
/*                                                                       
/* declaration		-> "cend"	$regexp;								<<< LexCEnd >>>
/*************************************************************************/
void LOADER::LexCEnd(const std::string & cend)
{
	generator->enableLex = true;

	// register line comment terminal
	int termId = AddTerminal(new PARAM(WORD, COMMENT_BLOCK_END));
	lex.RegisterComment(COMMENT_PRESET, termId, COMMENT_BLOCK_END, generator->terminal[termId].alias);

	// $cend is child of comment preset
	lex.genTerms[lex.genTermsNmb[termId]].preset = lex.presetsNmb[COMMENT_PRESET];

	// enable block comment generation
	lex.commentBlockEnd = termId;

	// line comment redefinition - is not error
	// modify comment's end regexp for searching
	lex.Init(termId, "{^}*(" + cend + ")", LEX::PRIORITY_NONE, Line(), GetError(), false);
	lex.Compile();
	if (lex.compilerState.value() != 0) {
		FatalError(MESSAGE_LDR1025);
		return ;
	}
}

/*************************************************************************/
/* LOADER::CheckPresetTerm()                                        
/*                                                                       
/* pr_com			-> param {CPT} "=" $regexp;							<<< LexPresetRegexp >>>
/*************************************************************************/
void LOADER::CPT(P_PARAM term)
{
	if (!term->IsTerminal()) {
		messageTable["name"] = term->value;
		FatalError(MESSAGE_LDR1026);
		return ;
	}

	PushPreset(PRESET_REGEXP);
}

/*************************************************************************/
/* LOADER::LexPresetRegexp()                                        
/*                                                                       
/* pr_com			-> param 	{CPT} "=" $regexp regexp_prior;	 <<< LexPresetRegexp >>>
/*************************************************************************/
void LOADER::LexPresetRegexp(P_PARAM term, const std::string & regexp, const string & prior)
{
	LexRegexp(term, regexp, prior);
	if (compilerState.is(CST_FATAL_ERROR)) {
		compilerState.set(CST_FATAL_ERROR);
		return ;
	}

	assert(generator->terminalNumber.find(term->value) != generator->terminalNumber.end());
	int termId = generator->terminalNumber[term->value];

	lex.genTerms[lex.genTermsNmb[termId]].preset = curPresetId;
}

/*************************************************************************/
/* LOADER::LexPresetTerm()                                          
/*                                                                       
/* pr_com			-> param;											<<< LexPresetTerm >>>
/*************************************************************************/
void LOADER::LexPresetTerm(P_PARAM term)
{
	if (!term->IsTerminal()) {
		FatalError(MESSAGE_LDR1027);
		return ;
	}

	int termId = AddTerminal(term);
	lex.genTerms[lex.genTermsNmb[termId]].preset = curPresetId;
}

/*************************************************************************/
/* LOADER::AddTerminal()                                                 */
/*                                                                       */
/*************************************************************************/
int LOADER::AddTerminal(P_PARAM terminal)
{
	// terminal must be terminal
	assert(terminal->IsTerminal());

	// add new terminal
	generator->AddTerminal(terminal->value);
	assert(generator->terminalNumber.find(terminal->value) != generator->terminalNumber.end());

	// register terminal in lex
	int termId = generator->terminalNumber[terminal->value];
	assert((size_t)termId < generator->terminal.size());
	lex.RegisterTerminal(termId, terminal->value, generator->terminal[termId].alias);

	// mark if terminal needs regexp
	if (terminal->fromWord) {
		needRegexp.insert(termId);
	}

	return termId;
}

/*************************************************************************/
/* LOADER::PTNM()                                                 
/*                                                                     
/* declaration		-> "preset"	$word {PTNM} "{" (pr_com *) {PTND} "}";
/*************************************************************************/
void LOADER::PTNM(const std::string & name)
{
	generator->enableLex = true;

	// all - reserved name for global terminals
	if (name == ALL_PRESET || name == COMMENT_PRESET) {
		messageTable["name"] = name;
		FatalError(MESSAGE_LDR1028);
		return ;
	}

	int presetId;
	if (!lex.AddPreset(name, presetId)) {
		messageTable["name"] = name;
		FatalError(MESSAGE_LDR1029);
		return ;
	}

	curPresetId = presetId;
}

/*************************************************************************/
/* LOADER::PTND()                                                   
/*                                                                       
/* declaration		-> "preset"	$word {PTNM} "{" (pr_com *) {PTND} "}";
/*************************************************************************/
void LOADER::PTND(const std::string & name)
{
	PopPreset();
}

/*************************************************************************/
/* LOADER::ESP()                                                    
/* 
/* declaration		-> "skip"	{ESP} "=" $regexp;						<<< LexSkip >>>
/*************************************************************************/
void LOADER::ESP(void)
{
	PushPreset(PRESET_REGEXP);
}

/*************************************************************************/
/* LOADER::Error()                                                       */
/*                                                                       */
/*************************************************************************/
void LOADER::Error(void)
{
	string token = "token \"" + TokenValue() + "\"";
	if (Token() == TOKEN_EOF) {
		token = "end of file";
	}
	messageTable["name"] = TokenValue();
	PrintMessage(MESSAGE_LDR1030);
}

/*************************************************************************/
/* LOADER::SPP()
/* 
/* declaration		-> alias_name 	  "=" $regexp [":" {SPP} $r_prior]; <<< LexRegexp >>>
/*************************************************************************/
void LOADER::SPP(void)
{
	PushPreset(PRESET_NAME);
}

/*************************************************************************/
/* LOADER::PreparePreset()
/* 
/* declaration		-> "preset" $name {PTNM} "{" (pr_com *) {PTND} "}";
/*************************************************************************/
void LOADER::PreparePreset(void)
{
	// set 'preset' preset for its commands
	PushPreset(PRESET_PRESET);

	// set 'name' preset for its name
	PushPreset(PRESET_NAME);
}

/*************************************************************************/
/* LOADER::EnableAutotype()                                              */
/*                                                                       */
/*************************************************************************/
void LOADER::EnableAutotype(void)
{
	enableAutotype = true;
}

/*************************************************************************/
/* LOADER::ENP()                                                         */
/*                                                                       */
/*************************************************************************/
void LOADER::ENP(void)
{
	PushPreset(PRESET_NAME);
}

/*************************************************************************/
/* LOADER::Message()                                                     */
/*                                                                       */
/*************************************************************************/
void LOADER::Message(const std::string & name, const std::string & format, const std::string & info)
{
	generator->enableMessages = true;

	if (generator->messages.find(name) != generator->messages.end()) {
		messageTable["name"] = name;
		FatalError(MESSAGE_LDR1033);
		return ;
	}

	generator->messages.insert(make_pair(name, MESSAGE(name, format, info)));
}

/*************************************************************************/
/* LOADER::StartNonterminal()                                            */
/*                                                                       */
/*************************************************************************/
void LOADER::StartNonterminal(const std::string & name)
{
	generator->startNonterm = name;
}

/*************************************************************************/
/* LOADER::FNT()                                                         */
/*                                                                       */
/*************************************************************************/
void LOADER::FNT(const std::string & name)
{
	if (generator->startNonterm.empty()) {
		generator->startNonterm = name;
	}
}

//
// End-of-file ECG_LOADER_IMPL.CPP
//