
/*************************************************************************/
/* MAIN.CPP                                                   05.11.2005 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "ecg_generator.h"

using namespace std;
using namespace ecg;

/*************************************************************************/
/* main()                                                                */
/*                                                                       */
/*************************************************************************/
int main(int argc, char * argv[])
{
	cout << "ECG v"
		<< (ECG_VERSION >> 16) << "."
		<< ((ECG_VERSION >> 8) & 0xff) << "."
		<< (ECG_VERSION & 0xff) << " (c) Evgeniy A. Cymbalyuk" << endl;
	cout << endl;

	if (argc == 1) {
		cout << "USE: ecg.exe options <input grammar>" << endl
			<< "Options:" << endl
			<< "     -o=<name>    	// implementation output filename" << endl
			<< "     -l=<name>    	// lexer output filename" << endl
			<< "     -m=<name>    	// messages output filename" << endl
			<< "     -i=<header>  	// include header for generated files" << endl
			<< "     -msg         	// specify this to build only messages" << endl;

		return 0;
	}

	GENERATOR gen;

	ifstream input;
	string outputName;
	string outputLexName;
	string outputMsgName;
	string implHeaderName;

	bool isMsgOnly = false;
	
	for (int i=1; i<argc; ++i) {
		if (strncmp(argv[i], "-o=", 3) == 0) {
			outputName = argv[i] + 3;
			continue;
		}
		if (strncmp(argv[i], "-l=", 3) == 0) {
			outputLexName = argv[i] + 3;
			continue;
		}
		if (strncmp(argv[i], "-i=", 3) == 0) {
			implHeaderName = argv[i] + 3;
			continue;
		}
		if (strncmp(argv[i], "-m=", 3) == 0) {
			outputMsgName = argv[i] + 3;
			continue;
		}
		if (strcmp(argv[i], "-msg") == 0) {
			isMsgOnly = true;
			continue;
		}
		if (argv[i][0] == '-') {
			cout << "ERROR: Unknown parameter '" << argv[i] << "'" << endl;
			return -1;
		}

		input.open(argv[i]);
		if (!input) {
			cout << "ERROR: Failed to open file '" << argv[i] << "'" << endl;
			return -1;
		}

		gen.Reset();

		gen.SetProcessMsgOnly(isMsgOnly);

		if (!gen.LoadGrammatics(input, cout)) {
			cout << "ERROR: Grammar loading failed" << endl;
			return -1;
		}

		if (!isMsgOnly) {
			if (!gen.CalcTable(cout)) {
				cout << "ERROR: Failed to create LALR table" << endl;
				return -1;
			}
		}

		string name = gen.GetImplName();
		assert(strlen(&name[0]) == name.size());
		strlwr(&name[0]);
		if (!outputName.empty()) {
			name = outputName;
		}

		string lexName = gen.GetName() + "_lex";
		assert(strlen(&lexName[0]) == lexName.size());
		strlwr(&lexName[0]);
		if (!outputLexName.empty()) {
			lexName = outputLexName;
		}

		string msgName = gen.GetName() + "_msg";
		assert(strlen(&msgName[0]) == msgName.size());
		strlwr(&msgName[0]);
		if (!outputMsgName.empty()) {
			msgName = outputMsgName;
		}

#define DECL_FILE(suffix, name, ext)	\
	string suffix##Name = name + ext; \
	ofstream suffix(suffix##Name.c_str()); \
	if (!suffix) { \
		cout << "ERROR: Failed to open output file '" << suffix##Name << "'" << endl; \
		return -1; \
	}

		// GENERATE COMPILER
		if (!isMsgOnly) {
			DECL_FILE(header, name, ".h");
			DECL_FILE(code,	name, ".cpp");
			DECL_FILE(hlp, name, ".txt");
			gen.SetInfo(headerName, codeName, implHeaderName);
			gen.Generate(header, code, hlp);
		}

		// GENERATE LEX
		if (!isMsgOnly && gen.IsLexEnabled()) {
			DECL_FILE(header, lexName, ".h");
			DECL_FILE(code, lexName, ".cpp");
			DECL_FILE(hlp, lexName, ".txt");
			gen.SetInfo(headerName, codeName, implHeaderName);
			gen.GenerateLex(header, code, hlp);
		}

		// GENERATE MSG
		if (gen.IsMessagesEnabled()) {
			DECL_FILE(header, msgName, ".h");
			DECL_FILE(code,	msgName, ".cpp");
			gen.SetInfo(headerName, codeName, implHeaderName);
			gen.GenerateMsg(header, code);
		}
	}

	cout << "DONE." << endl;

	return 0;
}

//
// End-of-file MAIN.CPP
//
